"""Graph represening a parsed kat script file."""

from enum import auto, Flag
import logging
import networkx as nx

LOGGER = logging.getLogger(__name__)

# All parsed nodes should descend from this root node. Each subsequent parsed file is
# appended to a new branch. This node is added by the compiler; having it automatically
# added in :meth:`.KatGraph.__init__` would interfere with generation of subgraphs.
ROOT_NODE_NAME = "kat"


class KatNodeType(Flag):
    ROOT = auto()

    # Containers.
    ELEMENT = auto()
    FUNCTION = auto()
    EXPRESSION = auto()
    GROUPED_EXPRESSION = auto()
    ARRAY = auto()
    NUMERICAL_ARRAY = auto()

    # Terminals.
    VALUE = auto()
    REFERENCE = auto()
    CONSTANT = auto()
    KEYWORD = auto()

    # Directive types.
    DIRECTIVE_NODES = ELEMENT | FUNCTION

    # Dependent types.
    DEPENDENT_NODES = REFERENCE

    # Nodes that don't have any further (incoming) dependencies. The compiler and
    # generator terminal nodes are slightly different due to the handling of references.
    COMPILER_TERMINAL_NODES = VALUE | KEYWORD
    GENERATOR_TERMINAL_NODES = VALUE | KEYWORD | CONSTANT | REFERENCE

    def __str__(self):
        return self.name


class KatEdgeType(Flag):
    ARGUMENT = auto()
    # Dependent parameter values/references.
    DEPENDENCY = auto()

    def __str__(self):
        return self.name


class KatGraph(nx.DiGraph):
    """Kat script graph."""

    def plot(self, *args, **kwargs):
        from ..plotting.graph import plot_graph

        plot_graph(self, *args, **kwargs)

    def is_tree(self):
        return nx.is_tree(self)

    def merge(self, other):
        """Merge `other` graph into this graph, rewriting paths in `other` if there are
        name collisions.

        Merging two identical graphs `kat --> kat.0 --> kat.0.1`, the result would be::

            kat --> kat.0 --> kat.0.1
                --> kat.1 --> kat.1.1

        Notes
        -----
        `self` and `other` must be trees (:meth:`.is_tree` returns True).
        """
        if not isinstance(other, type(self)):
            raise NotImplementedError(f"don't know how to merge from {type(other)}")

        LOGGER.debug(f"merging {repr(other)} into {repr(self)}")

        if not other.is_tree():
            raise RuntimeError("cannot merge syntax graph containing cycles")

        def _merge_branch(root, renames=None):
            if renames is None:
                renames = {}

            # Get the renamed root node, if necessary.
            final_root = renames.get(root, root)

            # The next highest order attribute of the root argument edges.
            if final_root in self.nodes:
                try:
                    maxorder = max(
                        self.dependent_arguments_edge_data(final_root, "order").values()
                    )
                except ValueError:
                    order = 0
                else:
                    order = maxorder + 1
            else:
                order = 0

                # Add the node.
                self.add_node(final_root, **other.nodes[root])

            for node, nodedata in other.sorted_dependent_argument_nodes(
                root, data=True
            ):
                final_node = other.item_node_name(order, final_root)
                if node != final_node:
                    renames[node] = final_node
                    LOGGER.debug(f"renaming {node} to {final_node}")

                # Merge edges.
                for target, _, edgedata in other.out_edges(node, data=True):
                    if other.path_parent(target) != root:
                        # The target is part of a different branch.
                        raise ValueError(f"{repr(other)} is not a tree")

                    assert "order" in edgedata
                    edgedata["order"] = order

                    self.add_edge(final_node, final_root, **edgedata)

                # Add the node and its children.
                self.add_node(final_node, **nodedata)
                _merge_branch(node, renames=renames)

                order += 1

        _merge_branch(ROOT_NODE_NAME)

    def argument(self, argument_node, adapter):
        """Get argument corresponding to node `argument_node`.

        This returns the argument object for a syntax graph node. It is useful for
        retrieving the names for positional arguments during parsing, but works for
        keyword arguments too.

        Returns
        -------
        :class:`.BoundArgument`
            The argument specified in the script corresponding to `argument_node`.

        Raises
        ------
        TypeError
            When `argument_node` corresponds to a positional argument that doesn't exist
            in the setter signature defined in `adapter`.
        """
        if key_token := self.nodes[argument_node].get("key_token"):
            # This is a keyword argument.
            name_or_index = key_token.raw_value
        else:
            # Probably a positional argument.
            name_or_index = self.argument_node_order(argument_node)

        return adapter.setter.bind_argument(name_or_index)

    def directive_graph(self, node):
        directive_nodes = self.dependent_argument_nodes(node)

        # Create a copy of the graph containing only the directive nodes and their
        # attributes.
        graph = self.subgraph(directive_nodes).copy()

        # Reassign parameter dependencies so the source is from the owning element.
        for target_directive, source_argument, data in self.edges(data=True):
            if data["type"] != KatEdgeType.DEPENDENCY:
                continue

            source_directive = self.branch_base(source_argument, node)
            graph.add_edge(target_directive, source_directive, **data)

        return graph

    def _node_view(self, nodes, data=False, default=None):
        """Return a node view in the same way that calling :class:`nx.DiGraph.nodes`
        would."""
        nodegraph = self.subgraph(nodes)
        if data is False:
            return nodegraph.nodes
        return nx.reportviews.NodeDataView(nodegraph.nodes, data, default)

    def _in_edge_view(self, edges, data=False, default=None):
        edgegraph = self.edge_subgraph(edges)
        if data is False:
            return edgegraph.edges
        return nx.reportviews.InEdgeView(edgegraph).data(data=data, default=default)

    def _out_edge_view(self, edges, data=False, default=None):
        edgegraph = self.edge_subgraph(edges)
        if data is False:
            return edgegraph.edges
        return nx.reportviews.OutEdgeView(edgegraph).data(data=data, default=default)

    def nodes_by_node_type(self, node_type, **kwargs):
        """Get nodes by type, with optional data."""
        return self._node_view(
            [
                node
                for node, ntype in self.nodes(data="type")
                if ntype and ntype in node_type
            ],
            **kwargs,
        )

    def in_edges_by_edge_type(self, node, edge_types, **kwargs):
        return self._in_edge_view(
            [
                (u, v)
                for u, v, edge_type in self.in_edges(node, data="type")
                if edge_type and edge_type in edge_types
            ],
            **kwargs,
        )

    def out_edges_by_edge_type(self, node, edge_types, **kwargs):
        return self._out_edge_view(
            [
                (u, v)
                for u, v, edge_type in self.out_edges(node, data="type")
                if edge_type and edge_type in edge_types
            ],
            **kwargs,
        )

    def in_edge_source_nodes_by_edge_type(self, node, edge_types, **kwargs):
        edges = self.in_edges_by_edge_type(node, edge_types)
        return self._node_view([edge[0] for edge in edges], **kwargs)

    def out_edge_target_nodes_by_edge_type(self, node, edge_types, **kwargs):
        edges = self.out_edges_by_edge_type(node, edge_types)
        return self._node_view([edge[1] for edge in edges], **kwargs)

    def dependent_argument_nodes(self, node, **kwargs):
        return self.in_edge_source_nodes_by_edge_type(
            node, KatEdgeType.ARGUMENT, **kwargs
        )

    def sorted_dependent_argument_nodes(self, node, **kwargs):
        def argument_order(node):
            # Deal with presence of data.
            if kwargs:
                node, _ = node
            return self.argument_node_order(node)

        return sorted(
            self.dependent_argument_nodes(node, **kwargs),
            key=argument_order,
        )

    def argument_node_order(self, node):
        """Get the `order` attribute of the edge linking argument `node` to its
        parent."""
        edges = self.out_edges_by_edge_type(node, KatEdgeType.ARGUMENT, data="order")
        if (nedges := len(edges)) != 1:
            raise RuntimeError(f"expected 1 argument edge, got {nedges}")
        _, __, order = next(iter(edges))
        return order

    def filter_argument_nodes(self, node, key):
        for argument_node, data in self.dependent_argument_nodes(node, data=True):
            if key(argument_node, data):
                yield argument_node

    def filter_dependent_nodes(self, node, key):
        nodes = self.out_edge_target_nodes_by_edge_type(
            node, KatEdgeType.DEPENDENCY, data=True
        )
        for dependent_node, data in nodes:
            if key(dependent_node, data):
                yield dependent_node

    def dependent_arguments_edge_data(self, parent, data):
        """Get mapping of dependent argument nodes to the value of the `data` attribute
        on the edge connecting them to `parent`."""
        edges_and_data = self.in_edges_by_edge_type(
            parent, KatEdgeType.ARGUMENT, data=data
        )
        return {target: d for target, _, d in edges_and_data}

    def is_independent(self, node):
        """Check if the node has no external dependencies.

        A node is independent if it is a terminal type or if its arguments have no non-
        argument incoming edges.
        """
        if dependencies := self.dependent_argument_nodes(node):
            return all(self.is_independent(dep) for dep in dependencies)
        node_type = self.nodes[node]["type"]
        if node_type not in KatNodeType.COMPILER_TERMINAL_NODES or self.in_edges(node):
            return False
        return True

    @classmethod
    def item_node_name(cls, name, parent_path):
        return f"{parent_path}.{name}"

    @classmethod
    def path_parent(cls, path):
        return ".".join(path.split(".")[:-1])

    @classmethod
    def branch_base(cls, path, reference):
        """The branch base node name for `path`, relative to `reference`."""
        prefix = f"{reference}."
        assert cls.is_subpath(path, prefix), f"{path} must start with {prefix}"
        path = path[len(prefix) :]
        return f"{prefix}{path.split('.')[0]}"

    @classmethod
    def is_subpath(cls, path, reference):
        """Check if `path` is a subpath of `reference`."""
        return path != reference and path.startswith(reference)

    def param_target_element_path(self, target, root_node):
        """The owning element path for `target`.

        Target should be in the form "element_name{.param_name{.param_name {...}}}.
        """
        # Grab the target's element name.
        pieces = target.split(".")
        assert len(pieces) >= 1, f"{target} is an invalid param path"
        owner = pieces[0]

        root_directive_nodes = self.dependent_argument_nodes(
            root_node, data="name_token"
        )

        for path, name_token in root_directive_nodes:
            if not name_token:
                # Not an element.
                continue
            if name_token.value == owner:
                return path

        raise ValueError(f"target '{target}' not found")



import pytest
from testutils.data import SI_PREFICES, INTEGERS, FLOATS, FLOATS_STD, IMAGINARIES
from testutils.tokens import (
    NAME,
    NUMBER,
    MINUS,
    LPAREN,
    RPAREN,
    EQUALS,
    SPACE,
    IMPLICITLINEEND,
    ENDMARKER,
)


@pytest.mark.parametrize(
    "string,expected",
    (
        ("0", [NUMBER(1, 1, "0"), IMPLICITLINEEND(1, 2), ENDMARKER(2)]),
        ("1", [NUMBER(1, 1, "1"), IMPLICITLINEEND(1, 2), ENDMARKER(2)]),
        ("10", [NUMBER(1, 1, "10"), IMPLICITLINEEND(1, 3), ENDMARKER(2)]),
        ("100", [NUMBER(1, 1, "100"), IMPLICITLINEEND(1, 4), ENDMARKER(2)]),
        ("12345", [NUMBER(1, 1, "12345"), IMPLICITLINEEND(1, 6), ENDMARKER(2)]),
        (
            "(-124561-1)",
            [
                LPAREN(1, 1),
                MINUS(1, 2),
                NUMBER(1, 3, "124561"),
                MINUS(1, 9),
                NUMBER(1, 10, "1"),
                RPAREN(1, 11),
                IMPLICITLINEEND(1, 12),
                ENDMARKER(2),
            ],
        ),
        # Long integer.
        (
            "a = 123141242151251616110",
            [
                NAME(1, 1, "a"),
                SPACE(1, 2),
                EQUALS(1, 3),
                SPACE(1, 4),
                NUMBER(1, 5, "123141242151251616110"),
                IMPLICITLINEEND(1, 26),
                ENDMARKER(2),
            ],
        ),
        # Separation using `_`.
        ("1_000", [NUMBER(1, 1, "1_000"), IMPLICITLINEEND(1, 6), ENDMARKER(2)]),
        ("1_0_00", [NUMBER(1, 1, "1_0_00"), IMPLICITLINEEND(1, 7), ENDMARKER(2)]),
        ("1_0_0_0", [NUMBER(1, 1, "1_0_0_0"), IMPLICITLINEEND(1, 8), ENDMARKER(2)]),
        ("1000_000", [NUMBER(1, 1, "1000_000"), IMPLICITLINEEND(1, 9), ENDMARKER(2)]),
        (
            "1_000_000",
            [NUMBER(1, 1, "1_000_000"), IMPLICITLINEEND(1, 10), ENDMARKER(2)],
        ),
    ),
)
def test_int(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected


@pytest.mark.parametrize(
    "string,expected",
    (
        ("3.14159", [NUMBER(1, 1, "3.14159"), IMPLICITLINEEND(1, 8), ENDMARKER(2)]),
        ("314159.", [NUMBER(1, 1, "314159."), IMPLICITLINEEND(1, 8), ENDMARKER(2)]),
        (".314159", [NUMBER(1, 1, ".314159"), IMPLICITLINEEND(1, 8), ENDMARKER(2)]),
        ("3e14159", [NUMBER(1, 1, "3e14159"), IMPLICITLINEEND(1, 8), ENDMARKER(2)]),
        ("3E123", [NUMBER(1, 1, "3E123"), IMPLICITLINEEND(1, 6), ENDMARKER(2)]),
        ("3e-1230", [NUMBER(1, 1, "3e-1230"), IMPLICITLINEEND(1, 8), ENDMARKER(2)]),
        ("3.14e159", [NUMBER(1, 1, "3.14e159"), IMPLICITLINEEND(1, 9), ENDMARKER(2)]),
        ("0.0", [NUMBER(1, 1, "0.0"), IMPLICITLINEEND(1, 4), ENDMARKER(2)]),
        ("0.00", [NUMBER(1, 1, "0.00"), IMPLICITLINEEND(1, 5), ENDMARKER(2)]),
        ("0.000", [NUMBER(1, 1, "0.000"), IMPLICITLINEEND(1, 6), ENDMARKER(2)]),
        ("1.0", [NUMBER(1, 1, "1.0"), IMPLICITLINEEND(1, 4), ENDMARKER(2)]),
        ("2.00", [NUMBER(1, 1, "2.00"), IMPLICITLINEEND(1, 5), ENDMARKER(2)]),
        ("3.000", [NUMBER(1, 1, "3.000"), IMPLICITLINEEND(1, 6), ENDMARKER(2)]),
        # Separation using `_`.
        ("1_000.", [NUMBER(1, 1, "1_000."), IMPLICITLINEEND(1, 7), ENDMARKER(2)]),
        ("1_0_00.", [NUMBER(1, 1, "1_0_00."), IMPLICITLINEEND(1, 8), ENDMARKER(2)]),
        ("1_0_0_0.", [NUMBER(1, 1, "1_0_0_0."), IMPLICITLINEEND(1, 9), ENDMARKER(2)]),
        (
            "1000_000.",
            [NUMBER(1, 1, "1000_000."), IMPLICITLINEEND(1, 10), ENDMARKER(2)],
        ),
        (
            "1_000_000.",
            [NUMBER(1, 1, "1_000_000."), IMPLICITLINEEND(1, 11), ENDMARKER(2)],
        ),
        (
            "1_234.5_678",
            [NUMBER(1, 1, "1_234.5_678"), IMPLICITLINEEND(1, 12), ENDMARKER(2)],
        ),
    ),
)
def test_float(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected


@pytest.mark.parametrize(
    "string,expected",
    (
        ("3.14159j", [NUMBER(1, 1, "3.14159j"), IMPLICITLINEEND(1, 9), ENDMARKER(2)]),
        ("314159.j", [NUMBER(1, 1, "314159.j"), IMPLICITLINEEND(1, 9), ENDMARKER(2)]),
        (".314159j", [NUMBER(1, 1, ".314159j"), IMPLICITLINEEND(1, 9), ENDMARKER(2)]),
        ("3e14159j", [NUMBER(1, 1, "3e14159j"), IMPLICITLINEEND(1, 9), ENDMARKER(2)]),
        ("3E123j", [NUMBER(1, 1, "3E123j"), IMPLICITLINEEND(1, 7), ENDMARKER(2)]),
        ("3e-1230j", [NUMBER(1, 1, "3e-1230j"), IMPLICITLINEEND(1, 9), ENDMARKER(2)]),
        (
            "3.14e159j",
            [NUMBER(1, 1, "3.14e159j"), IMPLICITLINEEND(1, 10), ENDMARKER(2)],
        ),
        # Separation using `_`.
        ("1_000j", [NUMBER(1, 1, "1_000j"), IMPLICITLINEEND(1, 7), ENDMARKER(2)]),
    ),
)
def test_imaginary(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected


@pytest.mark.parametrize("number,_a,_b", INTEGERS + FLOATS + IMAGINARIES)
def test_numbers(tokenizer, number, _a, _b):
    assert list(tokenizer.tokenize(number)) == [
        NUMBER(1, 1, number),
        IMPLICITLINEEND(1, 1 + len(str(number))),
        ENDMARKER(2),
    ]


@pytest.mark.parametrize("number,_a,_b", INTEGERS + FLOATS_STD)
@pytest.mark.parametrize("prefix,exponent", SI_PREFICES.items())
def test_numbers_with_si_prefices(tokenizer, number, _a, _b, prefix, exponent):
    string = f"{number}{prefix}"
    assert list(tokenizer.tokenize(string)) == [
        NUMBER(1, 1, string),
        IMPLICITLINEEND(1, 1 + len(str(string))),
        ENDMARKER(2),
    ]



"""Fix .so objects so that OSX codesign does not SIGKILL the process.

See issue https://gitlab.com/ifosim/finesse/finesse3/-/issues/460

We should not have to do this I feel...
"""
import glob
import shutil
import os

sos = glob.glob("**/*.so", recursive=True)

for so in sos:
    shutil.copyfile(so, so + "2")
    shutil.copyfile(so + "2", so)
    os.remove(so + "2")



"""
"""

"""Enum for the different types of modulator the Modulator component can represent.

am = Amplitude modulation
pm = Phase modulation
"""
cpdef enum ModulatorType:
    am
    pm
    # TODO ddb need to add in alignment and mode-matching modulator


"""Enum for the beamsplitter plane of incidence.

xz = x-z plane
yz = y-z plane
"""
cpdef enum PlaneOfIncidence:
    xz
    yz



import pytest
from finesse.symbols import OPERATORS, FUNCTIONS


# Override default element to take two arguments.
@pytest.fixture
def fake_element_cls(fake_element_cls):
    class FakeElement(fake_element_cls):
        def __init__(self, name, a=None, b=None):
            super().__init__(name)
            self.a = a
            self.b = b

    return FakeElement


@pytest.fixture
def fake_variadic_element_cls(fake_element_cls):
    class FakeVariadicElement(fake_element_cls):
        def __init__(self, name, a, *args, b=None, **kwargs):
            super().__init__(name)
            self.a = a
            self.args = args
            self.b = b
            self.kwargs = kwargs

    return FakeVariadicElement


@pytest.fixture
def spec(
    spec,
    set_spec_constructs,
    fake_element_adapter_factory,
    fake_element_cls,
    fake_variadic_element_cls,
):
    spec.register_element(
        fake_element_adapter_factory(
            fake_element_cls, full_name="fake_element", short_name="fake"
        )
    )
    spec.register_element(
        fake_element_adapter_factory(
            fake_variadic_element_cls,
            full_name="fake_variadic_element",
            short_name="fake_var",
        )
    )
    set_spec_constructs(
        "unary_operators",
        {"-": FUNCTIONS["neg"]},
        "binary_operators",
        {"+": OPERATORS["__add__"]},
        "expression_functions",
        {"cos": FUNCTIONS["cos"]},
    )

    return spec


@pytest.mark.parametrize("katdelim", (" ", "  ", "   "))
@pytest.mark.parametrize(
    "katargs,katkwargs",
    (
        ([], {}),
        ([""], {}),
        ([" "], {}),
        (["   "], {}),
        (["none"], {}),
        (["none", "none"], {}),
        (["none"], {"b": "none"}),
        ([], {"a": "none", "b": "none"}),
        (["1.23"], {}),
        (["1.23", "4.56"], {}),
        (["1.23"], {"b": "4.56"}),
        ([], {"a": "1.23", "b": "4.56"}),
    ),
)
def test_element(compiler, regenerate, katdelim, katargs, katkwargs):
    args = " ".join(katargs)
    kwargs = " ".join([f"{k}={v}" for k, v in katkwargs.items()])
    script = f"fake_element{katdelim}myelement{katdelim}{args}{katdelim}{kwargs}"
    model = compiler.compile(script)
    assert regenerate(model) == script



"""Collection of tools for computing different maps."""

import numpy as np
from scipy.special import erf


def make_coordinates(N, a):
    """Makes the 1D and square 2D grid of coordinates for map calculations.

    Parameters
    ----------
    N_samples : int
        Number of samples in each dimension (`N` x `N`)
    a : float
        Dimension of square grid (`a` x `a`) to generate

    Returns
    -------
    x,y : array_like[float]
        1D arrays for the x and y coordinates
    X,Y,R : array_like[float]
        2D Arrays of the X, Y, and radial R coordinates
    """

    x = y = np.linspace(-a, a, N)
    X, Y = np.meshgrid(x, y)
    R = np.sqrt(X**2 + Y**2)
    return x, y, X, Y, R


def circular_aperture(x, y, R_ap, x_offset=0, y_offset=0):
    """Circular aperture map.

    Parameters
    ----------
    x, y : array
        1D arrays describing uniform 2D grid to compute map
        over in meters
    R : float
        Radius of aperture in meters
    x_offset, y_offset : float, optional
        Offset of aperture from origin
    """
    X, Y = np.meshgrid(x, y)
    R = np.sqrt((X - x_offset) ** 2 + (Y - y_offset) ** 2)
    ap_map = np.ones_like(X)
    ap_map[R > R_ap] = 0
    return ap_map


def surface_point_absorber(
    xs, ys, w, h, power_absorbed, alpha=0.55e-6, kappa=1.38, zero_min=False
):
    """Models the surface deformation from a small point absorber in a coating of a
    mirror. It calcaultes the thermo-elastic deformation due to excess heat being
    deposited in the mirror.

    Parameters
    ----------
    xs, ys : array
        1D array for the x and y axis to calculate the distortion over
    w : double
        Area of the absorber size
    h : double
        Thickness of mirror
    power_absorbed : double
        Amount of power absorbed over the area w
    alpha : double, optional
        Thermo-elastic coefficient of material, default value for fused silica
    kappa : double, optional
        Thermal conductivity of the material, default value for fused silica

    Returns
    -------
    Height map in meters

    Notes
    -----
    Equation from:
        A. Brooks, et.al "Point absorbers in Advanced LIGO," Appl. Opt. (2021)
    """
    dx = xs[1] - xs[0]
    dy = ys[1] - ys[0]
    r = np.sqrt(np.add.outer(ys**2, xs**2))
    c0 = -1 / 2 - np.log((h**2 + np.sqrt(w**2 + h**2)) / w)
    out = np.zeros_like(r)
    # most of the time the absorber is very small, much smaller than
    # the discretisation of the map, so we can simplify this calculation
    if w > dx or w > dy:
        b1 = np.abs(r) <= w
        b2 = np.abs(r) > w
        out[b1] = -1 / 2 * (r[b1] / w) ** 2
        out[b2] = c0 + np.log((h**2 + np.sqrt(r[b2] ** 2 + h**2)) / r[b2])
    else:
        out = c0 + np.log((h**2 + np.sqrt(r**2 + h**2)) / r)

    if zero_min:
        out -= np.min(out)

    return alpha * power_absorbed / (2 * np.pi * kappa) * out


def overlap_piston_coefficient(x, y, z, weight_spot: float):
    """Computes the amount of weighted piston term there is in some 2D data like a
    surface map. This is computed by evaluating a weighted Hermite polynomial overlap
    integral in an efficient manner.

    Parameters
    ----------
    x, y : array_like
        1D Array of x and y describing the 2D plane of `z`
    z : array_like
        2D optical path depth [metres]
    weight_spot : float
        Beam spot size to weight over

    Returns
    -------
    piston : float
        amount of piston term
    """

    # TODO check why this the spot size is not adjusted here.
    # weight_spot /= np.sqrt(2)

    dx = (x[1] - x[0]) / weight_spot
    dy = (y[1] - y[0]) / weight_spot

    Wx = np.exp(-((x / weight_spot) ** 2))
    Wy = np.exp(-((y / weight_spot) ** 2))

    k00 = np.dot((z @ Wx), Wy) * dx / (np.sqrt(np.pi) * 2) * dy / (np.sqrt(np.pi) * 2)

    return 4 * k00


def overlap_tilt_coefficients(x, y, z, weight_spot: float):
    """Computes the amount of yaw and pitch terms present in a map's displacement data.

    This is computed by evaluating a weighted Hermite polynomial overlap integral in an
    efficient manner.

    Parameters
    ----------
    x, y : array_like
        1D Array of x and y describing the 2D plane of `z`
    z : array_like
        2D optical path depth [metres]
    weight_spot : float
        Beam spot size to weight over

    Returns
    -------
    k10, k01 : complex
        Complex-valued overlap tilt for the HG10 and HG01 mode
    """
    weight_spot /= np.sqrt(2)

    dx = (x[1] - x[0]) / weight_spot
    dy = (y[1] - y[0]) / weight_spot

    Wx = np.exp(-((x / weight_spot) ** 2))
    Wy = np.exp(-((y / weight_spot) ** 2))

    Hn = 2 * x / weight_spot * Wx
    Hm = 2 * y / weight_spot * Wy

    k10 = np.dot((z @ Hn), Wy) * dx / (np.sqrt(np.pi) * 2) * dy / (np.sqrt(np.pi) * 2)
    k01 = np.dot((z @ Wx), Hm) * dx / (np.sqrt(np.pi) * 2) * dy / (np.sqrt(np.pi) * 2)

    return (
        # convert back to normal x units in displacement map
        4 * k10 / weight_spot,
        4 * k01 / weight_spot,
    )


def overlap_curvature_coefficients(x, y, z, weight_spot: float):
    """Computes the amount of x and y curvature terms present in a map's displacement
    data.

    This is computed by evaluating a weighted Hermite polynomial overlap integral in an
    efficient manner.

    Parameters
    ----------
    x, y : array_like
        1D Array of x and y describing the 2D plane of `z`
    z : array_like
        2D optical path depth [metres]
    weight_spot : float
        Beam spot size to weight over

    Returns
    -------
    k20, k02 : complex
        Complex-valued overlap coefficients for the HG20 and HG02 mode
    """
    # Normalisation constant
    weight_spot /= np.sqrt(2)
    dx = (x[1] - x[0]) / weight_spot
    dy = (y[1] - y[0]) / weight_spot
    # Spot size weightings
    Wx = np.exp(-((x / weight_spot) ** 2))
    Wy = np.exp(-((y / weight_spot) ** 2))
    # 2nd order HG mode
    Hn = ((2 * x / weight_spot) ** 2 - 2) * Wx
    Hm = ((2 * y / weight_spot) ** 2 - 2) * Wy
    # Compute the overlap integrals
    k20 = np.dot((z @ Hn), Wy) * dx / (np.sqrt(np.pi) * 2**3) * dy / (np.sqrt(np.pi))
    k02 = np.dot((z @ Wx), Hm) * dx / (np.sqrt(np.pi) * 2**3) * dy / (np.sqrt(np.pi))
    return (
        # convert back to normal x units in displacement map
        4 * k20 / weight_spot / weight_spot,
        4 * k02 / weight_spot / weight_spot,
    )


def overlap_1D_curvature_coefficients(x, z, weight_spot: float):
    """Computes the amount of spot size weighted quadratic `x**2` term there is present
    in some 1D data.

    Parameters
    ----------
    x : ndarray(dtype=float)
        Sample points, ideally should be symmetric about 0.
    z : ndarray(dtype=float)
        function at sample points x
    weight_spot : float
        Spot size to use as weighting

    Returns
    -------
    quadratic_term : double
        Weighted quadratic term of z

    Notes
    -----
    This function is essentially evaluating a weighted Hermite polynomial
    overlap integral with `z(x)` to determine the linear term.

    .. math::

        \\int_{\\min(x)}^{\\max(x)} H_{2}(x) z(x) W(x) dx

    Where the weighting function is :math:`W(x) = exp(-x**2)`.
    """
    weight_spot /= np.sqrt(2)
    dx = (x[1] - x[0]) / weight_spot
    # exponential weighting of spot size
    W = np.exp(-((x / weight_spot) ** 2))
    # Second order Hermite, H_2(x)
    Hn = ((2 * x / weight_spot) ** 2 - 2) * W
    # normalisation constant as integral of H_0(x) over
    # domain is not 1
    norm = (
        0.5 * np.sqrt(np.pi) * (erf(x.max() / weight_spot) - erf(x.min() / weight_spot))
    )
    k20 = (z @ Hn).sum() * dx / norm
    return (
        # convert back to normal x units in displacement map
        k20
        / weight_spot
        / weight_spot
        / 2
    )


def overlap_1D_tilt_coefficients(x, z, weight_spot: float):
    """Computes the amount of spot size weighted linear `x` term there is present in
    some 1D data.

    Parameters
    ----------
    x : ndarray(dtype=float)
        Sample points, ideally should be symmetric about 0.
    z : ndarray(dtype=float)
        function at sample points x
    weight_spot : float
        Spot size to use as weighting

    Returns
    -------
    linear_term : double
        Weighted linear term of z

    Notes
    -----
    This function is essentially evaluating a weighted Hermite polynomial
    overlap integral with `z(x)` to determine the linear term.

    .. math::

        \\int_{\\min(x)}^{\\max(x)} H_{1}(x) z(x) W(x) dx

    Where the weighting function is :math:`W(x) = exp(-x**2)`.
    """
    weight_spot /= np.sqrt(2)
    dx = (x[1] - x[0]) / weight_spot
    # exponential weighting of spot size
    W = np.exp(-((x / weight_spot) ** 2))
    # Second order Hermite, H_1(x)
    Hn = 2 * x / weight_spot * W
    # normalisation constant as integral of H_0(x) over
    # domain is not 1
    norm = np.sqrt(np.pi) * (erf(x.max() / weight_spot) - erf(x.min() / weight_spot))
    k10 = (z @ Hn).sum() * dx / norm
    return (
        # convert back to normal x units in displacement map
        2
        * k10
        / weight_spot
    )


def rms(x, y, z, weight_spot: float, xo: float = 0, yo: float = 0):
    """Computes the spot weight RMS over some 2D data, such as optical path depth.

    Parameters
    ----------
    x, y : array_like
        1D Array of x and y describing the 2D plane of `z`
    z : array_like
        2D optical path depth [metres]
    weight_spot : float
        Beam spot size to weight over
    xo, yo : float
        Origin of the beam position

    Returns
    -------
    rms : float
        Root mean squared in units of whatever `z` is

    Notes
    -----
    Based on Equation 4 in:
        A. Brooks, et.al
        Overview of Advanced LIGO adaptive optics
        Appl. Opt. 55, 8256-8265 (2016)
    """
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    Wx = np.exp(-(((x - xo) / weight_spot) ** 2))
    Wy = np.exp(-(((y - yo) / weight_spot) ** 2))
    scaling = np.sqrt(2 / np.pi / weight_spot**2)
    deltaW = z
    deltaWbar = ((deltaW @ Wx) @ Wy) * dx * dy * scaling
    return np.sqrt((((deltaW - deltaWbar) ** 2 @ Wx) @ Wy) * dx * dy * scaling)


class BinaryReaderEOFException(Exception):
    def __init__(self):
        pass

    def __str__(self):
        return "Not enough bytes in file to satisfy read request"


class BinaryReader:
    """MetroPro binary data format reader."""

    # Map well-known type names into struct format characters.
    typeNames = {
        "int8": "b",
        "uint8": "B",
        "int16": "h",
        "uint16": "H",
        "int32": "i",
        "uint32": "I",
        "int64": "q",
        "uint64": "Q",
        "float": "f",
        "double": "d",
        "char": "s",
    }

    def __init__(self, fileName):
        self.file = open(fileName, "rb")

    def read(self, typeName, size=None):
        """Read a datatype from the binary file.

        Parameters
        ----------
        typename : str
            See `BinaryReader.typeNames`.
        size : int
            Number of bytes to read
        """
        import struct

        typeFormat = BinaryReader.typeNames[typeName.lower()]
        typeFormat = ">" + typeFormat
        typeSize = struct.calcsize(typeFormat)
        if size is None:
            value = self.file.read(typeSize)
            if typeSize != len(value):
                raise BinaryReaderEOFException
            unpacked = struct.unpack(typeFormat, value)[0]
        else:
            value = self.file.read(size * typeSize)
            if size * typeSize != len(value):
                raise BinaryReaderEOFException
            unpacked = np.zeros(size)
            for k in range(size):
                i = k * typeSize
                unpacked[k] = struct.unpack(typeFormat, value[i : i + typeSize])[0]
        return unpacked

    def seek(self, offset, refPos=0):
        """Offset in bytes and refPos gives reference position, where 0 means origin of
        the file, 1 uses current position and 2 uses the end of the file."""
        self.file.seek(offset, refPos)

    def __del__(self):
        self.file.close()


def read_metropro_file(filename):
    """Reading the metroPro binary data files. Translated from Hiro Yamamoto's
    'LoadMetroProData.m'.

    Parameters
    ----------
    filename
        Name of metropro data file.
    """
    f = BinaryReader(filename)
    # Read header
    hData = read_metropro_header(f)
    if hData["format"] < 0:
        print(
            "Error: Format unknown to readMetroProData()\nfilename: {:s}".format(
                filename
            )
        )
        return 0
    # Read phase map data
    # Skipping header and intensity data
    f.seek(hData["size"] + hData["intNBytes"])
    # Reading data
    dat = f.read("int32", size=hData["Nx"] * hData["Ny"])
    # Marking unmeasured data as NaN
    dat[dat >= hData["invalid"]] = np.nan
    # Scale data to meters
    dat = dat * hData["convFactor"]
    # Reshaping into Nx * Ny matrix
    dat = dat.reshape(hData["Ny"], hData["Nx"])
    # Flipping up/down, i.e., change direction of y-axis.
    dat = dat[::-1, :]
    # Auxiliary data to return
    # dxy = hData['cameraRes']
    # Unnecessary
    # x1 = dxy * np.arange( -(len(dat[0,:])-1)/2, (len(dat[0,:])-1)/2 + 1)
    # y1 = dxy * np.arange( -(len(dat[:,0])-1)/2, (len(dat[:,1])-1)/2 + 1)
    return dat, hData  # , x1, y1


def read_metropro_header(binary):
    """Reads header of the metroPro binary format files. Translated from the
    readerHeader() function within the 'LoadMetroProData.m' function written by Hiro
    Yamamoto.

    Parameters
    ----------
    binary : :class:`BinaryReader`
        Name of metropro data file.
    """
    f = binary
    hData = {}
    hData["magicNum"] = f.read("uint32")
    hData["format"] = f.read("int16")
    hData["size"] = f.read("int32")
    # Check if the magic string and format are known ones.
    if not (
        hData["format"] >= 1
        and hData["format"] <= 3
        and hData["magicNum"] - hData["format"] == int("881B036E", 16)
    ):
        hData["format"] = -1
    # Read necessary data
    hData["invalid"] = int("7FFFFFF8", 16)
    f.seek(60)
    # Intensity data, which we will skip over.
    hData["intNBytes"] = f.read("int32")
    # Top-left coordinates, which are useless.
    hData["X0"] = f.read("int16")
    hData["Y0"] = f.read("int16")
    # Number of data points alng x and y
    hData["Nx"] = f.read("int16")
    hData["Ny"] = f.read("int16")
    # Total data, 4*Nx*Ny
    hData["phaNBytes"] = f.read("int32")
    f.seek(218)
    # Scale factor determined by phase resolution tag
    phaseResTag = f.read("int16")
    if phaseResTag == 0:
        phaseResVal = 4096
    elif phaseResTag == 1:
        phaseResVal = 32768
    elif phaseResTag == 2:
        phaseResVal = 131072
    else:
        phaseResVal = 0
    f.seek(164)
    intf_scale_factor = f.read("float")
    hData["waveLength"] = f.read("float")
    f.seek(176)
    obliquity_factor = f.read("float")
    # Eq. in p12-6 in MetroPro Reference Guide
    hData["convFactor"] = (
        intf_scale_factor * obliquity_factor * hData["waveLength"] / phaseResVal
    )
    # Bin size of each measurement
    f.seek(184)
    hData["cameraRes"] = f.read("float")

    return hData



"""Expression unparsing tests."""

from finesse.script import KATSPEC, unparse
from finesse.symbols import simplification


def test_expression_references__unsimplified(model):
    """Test expressions containing references, without simplification.

    Without simplification (the default), Finesse will leave the expression for variable
    'c' in its original form, so the unparser will generate equivalent KatScript.
    """
    model.add_parameter("a", 1.0)
    model.add_parameter("b", 2.0)
    model.add_parameter(
        "c",
        (model.a.ref / model.b.ref + model.a.ref * model.b.ref + 2 - model.b.ref),
    )

    adapter = KATSPEC.elements["variable"]
    # Get the dump for the last variable.
    dump = list(adapter.getter(adapter, model))[2]
    script = unparse(dump)
    assert script == "variable c ((((a/b)+(a*b))+2)-b)"


def test_expression_references__simplified(model):
    """Test expressions containing references, with simplification.

    With simplification, Finesse will convert the expression for variable 'c' to a
    simplified form, and the unparser will generate a nested binary/unary expression
    from it.
    """
    model.add_parameter("a", 1.0)
    model.add_parameter("b", 2.0)

    with simplification():
        model.add_parameter(
            "c",
            (model.a.ref / model.b.ref + model.a.ref * model.b.ref + 2 - model.b.ref),
        )

    adapter = KATSPEC.elements["variable"]
    # Get the dump for the last variable.
    dump = list(adapter.getter(adapter, model))[2]
    script = unparse(dump)
    assert script == "variable c ((((-1*b)+2)+(a*(b**-1)))+(a*b))"



import logging
import numpy as np
cimport numpy as np

from libc.stdlib cimport free, calloc

from finesse.components.node import NodeType
from finesse.frequency cimport FrequencyContainer

LOGGER = logging.getLogger(__name__)


cdef class MatrixSystemWorkspaces:
    def __cinit__(self):
        self.ptr_to_refill = NULL
        self.ptr_to_rhs_refill = NULL
        self.ptr_to_noise_refill = NULL
        self.ptr_to_noise_input_refill = NULL
        self.ptr_noise_detectors = NULL

    def __init__(self):
        self.to_initial_fill = []
        self.to_refill = []
        self.to_rhs_refill = []
        self.to_noise_refill = []
        self.to_noise_input_refill = []
        self.noise_detectors = []

    def list_to_C(self):
        """Converts the python lists of workspaces into C Pyobject arrays for
        fast loop access.
        """
        if (
            self.ptr_to_refill != NULL
            or self.ptr_to_rhs_refill != NULL
            or self.ptr_to_noise_refill != NULL
            or self.ptr_to_noise_input_refill != NULL
        ):
            raise MemoryError()

        self.num_to_refill = len(self.to_refill)
        self.ptr_to_refill = <PyObject**> calloc(self.num_to_refill, sizeof(PyObject*))
        if not self.ptr_to_refill:
            raise MemoryError()
        cdef int i
        for i in range(self.num_to_refill):
            self.ptr_to_refill[i] = <PyObject*>self.to_refill[i]

        self.num_to_rhs_refill = len(self.to_rhs_refill)
        self.ptr_to_rhs_refill = <PyObject**> calloc(self.num_to_rhs_refill, sizeof(PyObject*))
        if not self.ptr_to_rhs_refill:
            raise MemoryError()
        for i in range(self.num_to_rhs_refill):
            self.ptr_to_rhs_refill[i] = <PyObject*>self.to_rhs_refill[i]

        self.num_to_noise_refill = len(self.to_noise_refill)
        self.ptr_to_noise_refill = <PyObject**> calloc(self.num_to_noise_refill, sizeof(PyObject*))
        if not self.ptr_to_noise_refill:
            raise MemoryError()
        for i in range(self.num_to_noise_refill):
            self.ptr_to_noise_refill[i] = <PyObject*>self.to_noise_refill[i]

        self.num_to_noise_input_refill = len(self.to_noise_input_refill)
        self.ptr_to_noise_input_refill = <PyObject**> calloc(self.num_to_noise_input_refill, sizeof(PyObject*))
        if not self.ptr_to_noise_input_refill:
            raise MemoryError()
        for i in range(self.num_to_noise_input_refill):
            self.ptr_to_noise_input_refill[i] = <PyObject*>self.to_noise_input_refill[i]

    def clear_workspaces(self):
        self.to_initial_fill.clear()
        self.to_refill.clear()
        self.to_rhs_refill.clear()
        self.to_noise_refill.clear()
        self.to_noise_input_refill.clear()
        self.noise_detectors.clear()

    def detector_list_to_C(self):
        self.num_noise_detectors = len(self.noise_detectors)
        if self.ptr_noise_detectors != NULL:
            raise MemoryError()
        self.ptr_noise_detectors = <PyObject**> calloc(self.num_noise_detectors, sizeof(PyObject*))
        if not self.ptr_noise_detectors:
            raise MemoryError()
        for i in range(self.num_noise_detectors):
            self.ptr_noise_detectors[i] = <PyObject*>self.noise_detectors[i]

    def __dealloc__(self):
        if self.ptr_to_refill:
            free(self.ptr_to_refill)

        if self.ptr_to_rhs_refill:
            free(self.ptr_to_rhs_refill)

        if self.ptr_to_noise_refill:
            free(self.ptr_to_noise_refill)

        if self.ptr_to_noise_input_refill:
            free(self.ptr_to_noise_input_refill)

        if self.ptr_noise_detectors:
            free(self.ptr_noise_detectors)


cdef class BaseSolver:
    """A linear set of systems can be represented as a matrix, each equation
    in this system is a particular state which we want to compute. The system
    is solved by applying some inputs into various states, or the right hand
    side (RHS) vector, and solving the system.

    The underlying matrix can be either a sparse or dense matrix. This class
    should not assume either, but merely call upon a standard matrix interface.
    Therefore the algorithm used for solving can vary significantly.
    The overall matrix is sectioned into submatricies which connect various
    states together.

    Nodes represent a physical location in the model in which some state of
    the system must be computed. Some nodes can have multiple states, such
    as multiple optical modes.
    """

    def __init__(
        self,
        str name,
        list nodes,
        FrequencyContainer optical_frequencies,
        dict signal_frequencies,
        bint is_signal_matrix,
        bint forced_refill,
        dict node_aliases,
        bint debug_mode = False,
    ):
        if is_signal_matrix:
            if signal_frequencies is None:
                raise Exception("Signal frequency containers not provided")
        else:
            if signal_frequencies is not None:
                raise Exception("Signal frequency container incorrectly provided for carrier simulation")

        self.is_signal_matrix = is_signal_matrix
        self.manual_rhs = False
        self.workspaces = MatrixSystemWorkspaces()
        self.forced_refill = forced_refill
        self.optical_frequencies = optical_frequencies
        self.signal_frequencies = signal_frequencies
        self.debug_mode = debug_mode

        if is_signal_matrix:
            # Get the unique FrequencyContainer objects for filling the info of.
            # This could be more optimal and dive into the references and check if
            # frequency containers are actually unique, but this mostly just stops
            # there being 100s of fsig containers from each node
            # TODO - optimise this away from a tuple
            self.unique_elec_mech_fcnts = tuple(set(signal_frequencies.values()))
            self.noise_sources = dict()

        self.any_frequencies_changing = False
        # Global flag for if any frequency is changing
        for _f in self.optical_frequencies.frequencies:
            if _f.symbol.is_changing:
                self.any_frequencies_changing = True
                break

        self.setup_nodes(nodes, node_aliases)

    def input_components(self):
        """Components that are injecting something into the simulation"""
        components = set()
        for ws in self.workspaces.to_rhs_refill:
            components.add(ws.owner)
        return components

    cpdef tuple get_node_frequencies(self, node):
        if node.type == NodeType.OPTICAL:
            return self.optical_frequencies.frequencies
        elif node.type == NodeType.MECHANICAL:
            return self.signal_frequencies[node].frequencies
        elif node.type == NodeType.ELECTRICAL:
            return self.signal_frequencies[node].frequencies
        else:
            raise ValueError()

    cpdef setup_nodes(self, list all_nodes, dict node_aliases):
        self.nodes = {n.full_name : n for n in all_nodes}
        self.num_nodes = len(self.nodes)
        self.node_aliases = {a.full_name: b.full_name for a, b, in node_aliases.items()}

        self.node_2_index = {}
        self.index_2_node = {}
        i = 0

        for n in self.nodes.values():
            if n in node_aliases:
                continue

            self.node_2_index[n.full_name] = i
            self.index_2_node[i] = n.full_name
            i += 1

        for n, a in node_aliases.items():
            # Map node n to node a details, essentially just an alias
            self.node_2_index[n.full_name] = self.node_2_index[a.full_name]

    cpdef assign_operators(self, connector_workspaces):
        """An important function. This takes all the connector workspaces - i.e. model elements
        that have requested some type of connection in the model - and ensures they have the
        correct submatrix allocated to them in for this solver.
        """
        raise NotImplementedError()

    cpdef assign_noise_operators(self, connector_workspaces):
        raise NotImplementedError()

    cpdef add_rhs(self, unicode key):
        raise NotImplementedError

    cpdef factor(self):
        raise NotImplementedError()

    cpdef refactor(self):
        raise NotImplementedError()

    cpdef solve(self):
        raise NotImplementedError()

    cpdef solve_noises(self):
        raise NotImplementedError()

    def initialise(self, sim):
        pass

    cpdef initial_fill(self):
        raise NotImplementedError()

    cpdef refill(self):
        raise NotImplementedError()

    cpdef refill_rhs(self):
        raise NotImplementedError()

    cpdef fill_rhs(self):
        raise NotImplementedError()

    cpdef fill_noise_inputs(self):
        raise NotImplementedError()

    cpdef construct(self):
        """This is called when workspaces and submatrices have been setup. Calling
        construct should now go and allocate the memory for the matrix and RHS.

        This method should be overwritten by an inheriting solver class with
        specfics of the solving technique.
        """
        raise NotImplementedError()

    cpdef destruct(self):
        """This is called when finishing and unbuilding the simulation.

        Classes that override this call should mindful of what this method is doing
        to and call it.
        """
        self.workspaces.clear_workspaces()

    cpdef initial_run(self):
        """Once a solver has been constructed it will most likely need to be initially
        filled and ran. Some sparse solvers for example must do a full factor first, then
        can perform faster refactors.

        This method should be overwritten by an inheriting solver class with
        specfics of the solving technique.
        """
        raise NotImplementedError()

    cpdef run(self):
        """Executes the simulation for model in its current state.

        Takes the following steps to compute an output:
         * If self.manual_rhs:
            * Clears the RHS vector
            * Fills the RHS vector
         * Fills the matrix
         * Solves
        """
        if not self.manual_rhs:
            self.clear_rhs()
            self.refill_rhs()
        self.refill()
        self.solve()
        if self.debug_mode:
            self.print_matrix()
            print("\nMatrix in dense format:\n")
            print(self._M.to_scipy_csr().todense())
            print("\nRight hand side vector:")
            self._M.print_rhs()

    def print_matrix(self):
        raise NotImplementedError()

    cpdef clear_rhs(self):
        raise NotImplementedError()

    cpdef Py_ssize_t findex(self, object node, Py_ssize_t freq):
        """
        Returns simulation unique index for a given frequency at this node.
        Used to refer to submatrices of HOMs in the interferometer matrix.

        Parameters
        ----------
        node : :class:`.Node`
            Node object to get the index of.
        freq : int
            Frequency index.

        Returns
        -------
        index : int
            Index of the `node` for a given frequency.
        """
        raise NotImplementedError()

    cpdef Py_ssize_t node_id(self, object node):
        if type(node) is str:
            return self.node_2_index[node]
        else:
            return self.node_2_index[node.full_name]

    cpdef get_node_info(self, node):
        """For a given node (object or name) the key parameters for where this node is
        represented in the matrix of linear equations.

        Parameters
        ----------
        node : [str | Node]
            The name or the Node object of the node.

        Returns
        -------
        dict: A dictionary containing the following information about the node:
            - index: The index of the node.
            - rhs_index: The index of the right-hand side vector associated with the node.
            - freq_index: The index of the frequency vector associated with the node.
            - nfreqs: The number of frequencies.
            - nhoms: The number of higher order modes. [TODO generalise to pixels/HOMs/whatever]
        """
        raise NotImplementedError()

    def get_frequency_object(self, frequency, node):
        """Get a :class:`.Frequency` object corresponding to a numerical or symbolic value.
        Returns none if nothing has been found.

        Parameters
        ----------
        f : number or :class:`.Symbol`
            Frequency to search for in this simulation.

        Returns
        -------
        :class:`.Frequency`
            The frequency object.
        """
        from finesse.symbols import Symbol

        if node.type == NodeType.OPTICAL:
            frequencies = self.optical_frequencies.frequencies
        elif node.type == NodeType.MECHANICAL:
            if not self.is_signal_matrix:
                return None
            frequencies = self.signal_frequencies[node].frequencies
        elif node.type == NodeType.ELECTRICAL:
            if not self.is_signal_matrix:
                return None
            frequencies = self.signal_frequencies[node].frequencies

        if isinstance(frequency, Symbol):
            if frequency.is_changing:
                # if it's tunable we want to look for the symbol that is just this
                # lasers frequency, as it will be changing
                for f in frequencies:
                    if f.symbol == frequency:
                        return f

        f_value = float(frequency)
        # otherwise do some value comparisons
        for f in frequencies:
            if np.isclose(float(f.f), f_value, atol=1e-15, rtol=1e-15):
                return f

        return None

    cpdef update_frequency_info(self):
        self.optical_frequencies.update_frequency_info()
        if self.is_signal_matrix:
            for i in range(len(self.unique_elec_mech_fcnts)):
                (<FrequencyContainer>self.unique_elec_mech_fcnts[i]).update_frequency_info()



import pytest
import finesse
import numpy as np
from scipy.special import erf


@pytest.fixture
def dc_xysplit_model():
    model = finesse.Model()
    model.parse(
        """
    l l1
    bs bs1 R=1 T=0 ybeta=bs1.xbeta
    nothing n1
    pd X n1.p1.i pdtype=xsplit
    pd Y n1.p1.i pdtype=ysplit
    link(l1, bs1, 10, n1)
    modes(maxtem=1)
    gauss g1 bs1.p2.o w0=1m z=0
    xaxis(bs1.xbeta, lin, -10u, 10u, 10)
    """
    )
    return model


def test_dc_xysplit(dc_xysplit_model):
    model = dc_xysplit_model
    model.run()


@pytest.mark.parametrize("D", [1, 10, 100])
@pytest.mark.parametrize("a,b,sign", [("x", "y", +1), ("y", "x", -1)])
def test_pd_split(D, a, b, sign):
    model = finesse.Model()
    model.parse(
        f"""
    l l1
    bs bs1 R=1 T=0 {b}beta=-bs1.{a}beta # Steering mirror
    nothing n1
    pd qpdx n1.p1.i pdtype=xsplit
    pd qpdy n1.p1.i pdtype=ysplit
    link(l1, bs1, {D}, n1)

    modes(maxtem=1)
    gauss g1 bs1.p2.o w0=1m z=0
    xaxis(bs1.{a}beta, lin, -1n, 1n, 1)
    """
    )

    out = model.run()
    delta = out.x1 * D
    err_x = out["qpdx"] + sign * erf(2 * np.sqrt(2) * delta / model.n1.p1.i.qx.w)
    err_y = out["qpdy"] - sign * erf(2 * np.sqrt(2) * delta / model.n1.p1.i.qx.w)

    assert abs(err_x).max() < 1e-15
    assert abs(err_y).max() < 1e-15



#!/usr/bin/env python

from __future__ import annotations

try:
    import tomllib

    # tomllib not available in python < 3.11
except ImportError:
    try:
        import tomli as tomllib
    except ImportError as e:
        raise ImportError("This pre-commit hook requires tomli on python < 3.11") from e

import re
from pathlib import Path

ROOT = Path(__file__).absolute().parent.parent
PYPROJECT_TOML = ROOT / "pyproject.toml"
ENVIRONMENT_YML = ROOT / "environment.yml"
ENVIRONMENT_WIN_YML = ROOT / "environment-win.yml"


class VersionMismatchError(ValueError):
    "Package mentioned twice with different versions"


class BuildReqMismatchError(ValueError):
    "Build requirements and 'inplacebuild' requirements not identical"


def parse_req_line(line: str) -> tuple[str, str]:
    """Parse a requirement spec into package name and requirement.

    "numpy > 1" -> ("numpy", ">1")

    Parameters
    ----------
    line : str
        Line from a requirement file, e.g. "numpy > 1"

    Returns
    -------
    tuple[str, str]
        Name, Requirement
    """
    line = re.sub(r"\s", repl="", string=line)
    # see https://peps.python.org/pep-0508/#names
    match = re.match(
        r"""(^
            [A-Z0-9] # start with at least 1 alpha-numerical character
            [A-Z0-9._-]* # any number of alphanumericals, dots, underscores or hyphens
            [A-Z0-9] # Final character can not be a dot, underscore or hyphen
            )
            (.*) # followed by possible requirements specifier
            """,
        line,
        flags=re.I | re.VERBOSE,
    )
    assert match is not None
    name = match.groups()[0].lower()
    if match.groups()[1]:
        req = match.groups()[1].lower()
    else:
        req = ""
    return name, req


def check_version_mismatch(deps: list[str], source: Path) -> None:
    versions = {}
    for dep in deps:
        name, req = parse_req_line(dep)
        if name in versions and versions[name] != req:
            raise VersionMismatchError(
                f"Dependency {name} specified twice with different versions in {source}"
                f"\n Version 1: '{req}'"
                f"\n Version 2: '{versions[name]}'"
            )
        else:
            versions[name] = req


def check_pyproject_toml() -> set[str]:
    """Checks whether the build requirements in the pyproject.toml are identical to the
    'inplacebuild' optional dependencies, and whether there are any packages listed
    twice with a different version requirement.

    Returns
    -------
    set[str]
        Set of all python requirements in the pyproject.toml file

    Raises
    ------
    ValueError
        When there is a mismatch between build requirements and 'inplacebuild'
        requirements
    """
    with open(PYPROJECT_TOML, "rb") as f:
        cfg = tomllib.load(f)
    build_system = set(cfg["build-system"]["requires"])
    inplacebuild = set(cfg["project"]["optional-dependencies"]["inplacebuild"])
    mismatch = build_system.symmetric_difference(inplacebuild)
    if mismatch:
        raise BuildReqMismatchError(
            f"Mismatch between build system requirements in pyproject.toml "
            f"and 'inplacebuild' optional requirements, "
            f" symmetric difference: {mismatch}"
        )
    optional_deps = sum(cfg["project"]["optional-dependencies"].values(), start=[])
    all_deps = cfg["project"]["dependencies"] + optional_deps
    check_version_mismatch(all_deps, PYPROJECT_TOML)
    return set(all_deps)


def main():
    check_pyproject_toml()


if __name__ == "__main__":
    main()



# %%
import finesse
from finesse.analysis.actions import (
    DragLocks,
    OptimiseRFReadoutPhaseDC,
    SetLockGains,
    RunLocks,
    Series,
    Maximize,
    Change,
    Noxaxis,
)
import pytest

finesse.configure(plotting=True)


@pytest.fixture
def locks_model():
    kat = finesse.Model()
    kat.parse(
        """
        l L0 P=1
        s l_mod1 L0.p1 eo1.p1
        mod eo1 10M 0.1

        s s0 eo1.p2 BS.p1
        bs BS R=0.5 T=0.5

        s s1 BS.p2 NI.p1
        m NI R=0.99 T=0.01 Rc=1429 phi=90
        s CAV NI.p2 NE.p1 L=10
        m NE R=0.991 T=0.009 Rc=1430 phi=90

        s s2 BS.p3 EI.p1
        m EI R=0.99 T=0.01 Rc=1429 phi=0
        s CAV2 EI.p2 EE.p1 L=10
        m EE R=0.991 T=0.009 Rc=1430 phi=0

        dof NEz NE.dofs.z +1
        dof EEz EE.dofs.z +1
        dof NIz NI.dofs.z +1

        readout_rf rd_pdh1 NI.p1.o f=10M
        readout_rf rd_pdh2 EI.p1.o f=10M
        readout_rf rd_DF BS.p4.o f=10M

        lock cav1_lock rd_pdh1.outputs.I NEz.DC 1 1e-9
        lock cav2_lock rd_pdh2.outputs.I EEz.DC 1 1e-9
        lock DF_lock rd_DF.outputs.I NIz.DC 1 1e-9

        cav cav1 NI.p2.o
        cav cav2 EI.p2.o
        """
    )
    return kat


def test_run_locks_newton_method(locks_model):
    sol = locks_model.run(RunLocks(method="newton", display_progress=False))
    assert sol.iters < 10


def test_drag_locks_newton_method(locks_model):
    sol = locks_model.run(
        DragLocks(
            method="newton",
            parameters=["BS.xbeta"],
            stop_points=[2e-6],
            display_progress=False,
            relative=True,
        )
    )
    assert sol.iters < 10


def test_pdh_lock_optimisation_action():
    model = finesse.Model()
    model.parse(
        """
    l l1
    mod mod1 10M 0.1 mod_type=pm
    readout_rf PD f=mod1.f phase=37.3 output_detectors=True optical_node=m1.p1.o
    m m1 R=0.9999 T=0.0001
    m m2 R=1 T=0
    link(l1, mod1, m1, 1, m2)
    dof CAV m2.phi +1
    lock cav_lock PD_I CAV.DC 0.01 1e-3

    pd Pcirc m2.p1.o
    ad Ecirc m2.p1.o f=0
    ad Ein m1.p1.i f=0
    """
    )

    test_lock = Series(
        Maximize(model.Pcirc, model.CAV.DC),
        OptimiseRFReadoutPhaseDC(),
        SetLockGains(),
        Change({model.CAV.DC: 0.1}, relative=True),
        RunLocks(),
        Noxaxis(),
    )
    sol = model.run(test_lock)
    g = 4 / model.m1.T  # approximate cavity gain for high finesse
    # model.CAV.DC, model.PD.phase, model.cav_lock.gain, abs(sol['noxaxis']['Ein'])**2 * g, sol['noxaxis']['Pcirc']
    assert abs(abs(sol["noxaxis"]["Ein"]) ** 2 * g - sol["noxaxis"]["Pcirc"]) < 3



cdef extern from "math.h" nogil:
    double atan2(double y, double x)
    double cos(double arg)
    double exp(double arg)
    double fabs(double arg)
    double fmax(double x, double y)
    double log(double x)
    double sin(double arg)
    double sqrt(double arg)
    double asin(double arg)
    double acos(double arg)
    double cosh(double arg)
    double sinh(double arg)
    double ceil(double arg)
    double floor(double arg)
    int isfinite(float arg)
    float NAN
    float INF

cdef extern from "constants.h":
    long double PI
    double DEG2RAD
    double RAD2DEG
    double ROOT2

## TODO (sjr) Used the fused type below on some functions
##            in this file for template-like behaviour

#ctypedef fused real_numeric_t: # non-complex numeric type
#    short
#    int
#    long
#    float
#    double

cpdef inline double degrees(double x) noexcept nogil:
    return x * RAD2DEG


cpdef inline double radians(double x) noexcept nogil:
    return x * DEG2RAD

cpdef inline int sgn(double x) noexcept nogil:
    return (0.0 < x) - (x < 0.0)

cpdef inline double msign(int n) noexcept nogil:
    return -1.0 if n % 2 else 1.0


cpdef inline int nmin(int n, int m) noexcept nogil:
    return n if n < m else m


cpdef inline int nmax(int n, int m) noexcept nogil:
    return n if n > m else m


cpdef inline bint float_eq(double x, double y) noexcept nogil:
    if x == 0.0 and y == 0.0:
        return 1
    else:
        return fabs(x - y) / fmax(fabs(x), fabs(y)) < 1e-13


cpdef double factorial(int n) noexcept nogil
cpdef double sqrt_factorial(int n) noexcept nogil

cpdef double hermite(int n, double x) noexcept nogil



from networkx.readwrite import json_graph

import finesse


def test_symbolic_network():
    model = finesse.Model(
        """
    var Y 0.5
    var X Y
    var Z 10
    m m1 R=1-m1.T T=X phi=X+Z
    """
    )
    REF_GRAPH = {
        "directed": True,
        "multigraph": False,
        "graph": [],
        "nodes": [
            {"id": model.X},
            {"id": model.m1.phi},
            {"id": model.Z},
            {"id": model.m1.R},
            {"id": model.m1.L},
            {"id": model.m1.T},
            {"id": model.Y},
        ],
        "adjacency": [
            [{"id": model.m1.phi}, {"id": model.m1.T}],
            [],
            [{"id": model.m1.phi}],
            [{"id": model.m1.L}],
            [],
            [{"id": model.m1.L}, {"id": model.m1.R}],
            [{"id": model.X}],
        ],
    }
    assert json_graph.adjacency_data(model.symbolic_network) == REF_GRAPH



from difflib import unified_diff

import pytest

from finesse import Model
from finesse.components.replace import MissingReplacementError


def compare_kat(a: str, b: str):
    if a != b:
        raise AssertionError("\n".join(unified_diff(a.splitlines(), b.splitlines())))


def test_explicit_space():
    # CASE 1: simple linear model using explicit spaces

    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
m m1
s s2 m1.p2 m2.p1
m m2"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        verbose=False,
    )
    expected = """\
l l1
s s1 l1.p1 m_miron.p1
m m_miron R=0.5 T=0.5 L=0.0
s s2 m_miron.p2 m2.p1
m m2 R=0.5 T=0.5 L=0.0"""

    compare_kat(new_kat, expected)


def test_explicit__leading_space():
    # CASE 1: simple linear model using explicit spaces

    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
    m m1
s s2 m1.p2 m2.p1
m m2"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        verbose=False,
    )
    expected = """\
l l1
s s1 l1.p1 m_miron.p1
m m_miron R=0.5 T=0.5 L=0.0
s s2 m_miron.p2 m2.p1
m m2 R=0.5 T=0.5 L=0.0"""

    compare_kat(new_kat, expected)


def test_optical_ports_too_short():
    # CASE 1: simple linear model using explicit spaces

    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
m m1
s s2 m1.p2 m2.p1
m m2"""
    )

    with pytest.raises(MissingReplacementError):
        model.m1.replace(
            "m m_miron R=0.5 T=0.5 L=0.0",
            optical_ports=[
                "m_miron.p1",
            ],
            verbose=False,
        )


def test_optical_ports_none():
    # CASE 1: simple linear model using explicit spaces

    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
m m1
s s2 m1.p2 m2.p1
m m2"""
    )

    with pytest.raises(MissingReplacementError):
        model.m1.replace(
            "m m_miron R=0.5 T=0.5 L=0.0",
            optical_ports=None,
            verbose=False,
        )


def test_optical_ports_none_allowed():
    model = Model("m m1 R=0.5 T=0.5 L=0.0")
    new_kat = model.m1.replace("m m_miron R=0.5 T=0.5 L=0.0", verbose=False)
    assert new_kat == "m m_miron R=0.5 T=0.5 L=0.0"


# CASE 2: using mirror aliases
def test_aliases():
    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
mirror m1 R=0.5 T=0.5 L=0.0
s s2 m1.p2 m2.p1
m m2 R=0.5 T=0.5 L=0.0"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        verbose=False,
    )
    expected = """\
l l1
s s1 l1.p1 m_miron.p1
m m_miron R=0.5 T=0.5 L=0.0
s s2 m_miron.p2 m2.p1
m m2 R=0.5 T=0.5 L=0.0"""

    compare_kat(new_kat, expected)


# CASE 3: using arbitrary whitespace in katscript (seems to be removed by the unparsers?)
def test_whitespace():
    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
mirror           m1 R=0.5 T=0.5 L=0.0
s s2 m1.p2 m2.p1
m m2 R=0.5 T=0.5 L=0.0"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        verbose=False,
    )
    expected = """\
l l1
s s1 l1.p1 m_miron.p1
m m_miron R=0.5 T=0.5 L=0.0
s s2 m_miron.p2 m2.p1
m m2 R=0.5 T=0.5 L=0.0"""

    compare_kat(new_kat, expected)


# CASE 4: replacing an end mirror
def test_single_connection():
    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
mirror m1 R=0.5 T=0.5 L=0.0"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        verbose=False,
    )
    expected = """\
l l1
s s1 l1.p1 m_miron.p1
m m_miron R=0.5 T=0.5 L=0.0"""

    compare_kat(new_kat, expected)


# CASE 5: using the 'link' command with explicit ports
# unparsers turns links into spaces
def test_link_explicit():
    model = Model(
        """\
l l1
link(l1.p1, m1.p1)
mirror m1 R=0.5 T=0.5 L=0.0
link(m1.p2, m2.p1)
m m2 R=0.5 T=0.5 L=0.0"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        verbose=False,
    )
    expected = """\
l l1

m m_miron R=0.5 T=0.5 L=0.0

m m2 R=0.5 T=0.5 L=0.0

space l1_p1__m1_p1 portA=l1.p1 portB=m_miron.p1
space m1_p2__m2_p1 portA=m_miron.p2 portB=m2.p1"""

    compare_kat(new_kat, expected)


# CASE 6: using the 'link' command without explicit ports
def test_link_implicit():
    model = Model(
        """\
l l1
link(l1, m1)
mirror m1 R=0.5 T=0.5 L=0.0
link(m1, m2)
m m2 R=0.5 T=0.5 L=0.0"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        verbose=False,
    )
    expected = """\
l l1

m m_miron R=0.5 T=0.5 L=0.0

m m2 R=0.5 T=0.5 L=0.0

space l1_p1__m1_p1 portA=l1.p1 portB=m_miron.p1
space m1_p2__m2_p1 portA=m_miron.p2 portB=m2.p1"""

    compare_kat(new_kat, expected)


def test_detectors():
    model = Model(
        """\
m m1 R=0.5 T=0.5 L=0.0
ad amp1 m1.p1.i f=0
ad amp2 m1.p1.o f=0
pd pow1 m1.p2.i
pd pow2 m1.p2.o
"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        verbose=False,
    )
    expected = """\
m m_miron R=0.5 T=0.5 L=0.0
ad amp1 m_miron.p1.i f=0
ad amp2 m_miron.p1.o f=0
pd pow1 m_miron.p2.i
pd pow2 m_miron.p2.o"""

    compare_kat(new_kat, expected)


def test_component():
    model = Model(
        """\
m m1 R=0.5 T=0.5 L=0.0
pendulum pend2 m1"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        component="m_miron",
        optical_ports=None,
        verbose=False,
    )

    expected = """\
m m_miron R=0.5 T=0.5 L=0.0
pendulum pend2 m_miron"""

    compare_kat(new_kat, expected)


def test_component_spaces():
    model = Model(
        """\
m m1 R=0.5 T=0.5 L=0.0
pendulum pend2 m1 mass=4"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        component="m_miron",
        optical_ports=None,
        verbose=False,
    )

    expected = """\
m m_miron R=0.5 T=0.5 L=0.0
pendulum pend2 m_miron mass=4"""

    compare_kat(new_kat, expected)


def test_component_comment():
    model = Model(
        """\
m m1 R=0.5 T=0.5 L=0.0
pendulum pend2 m1#comment"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        component="m_miron",
        optical_ports=None,
        verbose=False,
    )

    expected = """\
m m_miron R=0.5 T=0.5 L=0.0
pendulum pend2 m_miron#comment"""

    compare_kat(new_kat, expected)


def test_component_keyword():
    model = Model(
        """\
m m1 R=0.5 T=0.5 L=0.0
pendulum pend2 connected_to=m1 mass=4"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        component="m_miron",
        optical_ports=None,
        verbose=False,
    )

    expected = """\
m m_miron R=0.5 T=0.5 L=0.0
pendulum pend2 connected_to=m_miron mass=4"""

    compare_kat(new_kat, expected)


def test_component_keyword_comment():
    model = Model(
        """\
m m1 R=0.5 T=0.5 L=0.0
pendulum pend2 connected_to=m1#comment"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        component="m_miron",
        optical_ports=None,
        verbose=False,
    )

    expected = """\
m m_miron R=0.5 T=0.5 L=0.0
pendulum pend2 connected_to=m_miron#comment"""

    compare_kat(new_kat, expected)


def test_mechanical_ports():
    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
m m1
s s2 m1.p2 m2.p1
m m2
ad motion m1.mech.z f=0
"""
    )

    new_kat = model.m1.replace(
        "m m_miron R=0.5 T=0.5 L=0.0",
        optical_ports=["m_miron.p1", "m_miron.p2"],
        mechanical_ports=["m_miron.mech"],
        verbose=False,
    )

    expected = """\
l l1
s s1 l1.p1 m_miron.p1
m m_miron R=0.5 T=0.5 L=0.0
s s2 m_miron.p2 m2.p1
m m2 R=0.5 T=0.5 L=0.0
ad motion m_miron.mech.z f=0"""

    compare_kat(new_kat, expected)


def test_mechanical_ports_missing():
    model = Model(
        """\
l l1
s s1 l1.p1 m1.p1
m m1
s s2 m1.p2 m2.p1
m m2
ad motion m1.mech.z f=0
"""
    )

    with pytest.raises(MissingReplacementError):
        model.m1.replace(
            "m m_miron R=0.5 T=0.5 L=0.0",
            optical_ports=["m_miron.p1", "m_miron.p2"],
            mechanical_ports=None,
            verbose=False,
        )


def test_complicated():
    model = Model(
        """\
l l1

s LA l1.p1 PLACEHOLDER.p1
nothing PLACEHOLDER
s LB PLACEHOLDER.p2 central_bs.p1

bs central_bs R=0.5 T=0.5
s LW central_bs.p2 WE.p1 L=100
m WE R=0.9 T=0.1
s LN central_bs.p3 NE.p1 L=100
m NE R=0.9 T=0.1
pd pd1 central_bs.p4.o"""
    )

    IMC_KATSCRIPT = """\
bs bs1 R=0.9 T=0.1
s s1 bs1.p2 bs2.p1
bs bs2 R=0.9 T=0.1
s s2 bs2.p2 bs3.p2
bs bs3 R=0.9 T=0.1
s s3 bs1.p3 bs3.p1
"""

    new_kat = model.PLACEHOLDER.replace(
        IMC_KATSCRIPT, optical_ports=["bs1.p1", "bs3.p3"]
    )

    expected = """\
l l1

s LA l1.p1 bs1.p1
bs bs1 R=0.9 T=0.1
s s1 bs1.p2 bs2.p1
bs bs2 R=0.9 T=0.1
s s2 bs2.p2 bs3.p2
bs bs3 R=0.9 T=0.1
s s3 bs1.p3 bs3.p1
s LB bs3.p3 central_bs.p1

bs central_bs R=0.5 T=0.5 L=(1-(central_bs.R+central_bs.T))
s LW central_bs.p2 WE.p1 L=100
m WE R=0.9 T=0.1 L=(1-(WE.R+WE.T))
s LN central_bs.p3 NE.p1 L=100
m NE R=0.9 T=0.1 L=(1-(NE.R+NE.T))
pd pd1 central_bs.p4.o"""

    compare_kat(new_kat, expected)



"""Single-frequency array of complex amplitudes detector."""

import logging

import numpy as np

from finesse import BeamParam
from finesse.components.node import Node, NodeType
from finesse.detectors.general import Detector
from finesse.exceptions import ConvergenceException, FinesseException
from finesse.parameter import float_parameter, bool_parameter
from finesse.detectors.workspace import DetectorWorkspace
from finesse.gaussian import optimise_HG00_q_scipy

LOGGER = logging.getLogger(__name__)


class OptimalQWorkspace(DetectorWorkspace):
    """Workspace for calculating the output of the optimial beamparameter (q) detectors.

    Parameters
    ----------
    owner : OptimalQ
        Detector which owns this workspace
    sim
        Simulation this workspace should be created for
    """

    def __init__(self, owner, sim):
        needs_carrier = False
        needs_signal = False
        self.is_f_changing = owner.f.is_changing
        if owner.f.eval() is None:
            raise ValueError(
                f"{owner.f}: frequency value is `None`, check values have been set correctly."
            )
        fval = float(owner.f)
        fs = []

        if sim.carrier:
            f = sim.carrier.get_frequency_object(fval, owner.node)
            if f is not None:
                needs_carrier = True
                fs.append((f, sim.carrier))

        if sim.signal:
            f = sim.signal.get_frequency_object(fval, owner.node)
            if f is not None:
                needs_signal = True
                fs.append((f, sim.signal))

        if len(fs) == 0:
            raise Exception(
                f"Error in OptimalQ detector {owner.name}:\n"
                f"    Could not find a frequency bin at {owner.f}"
            )
        elif len(fs) > 1:
            raise Exception(
                f"Error in OptimalQ detector {owner.name}:\n"
                f"    Found multiple frequency bins at {owner.f}"
            )

        super().__init__(
            owner, sim, needs_carrier=needs_carrier, needs_signal=needs_signal
        )
        freq, self.mtx = fs[0]
        self.node_idx = self.mtx.node_id(owner.node)
        self.f_idx = freq.index

        self.set_output_fn(self.__output)
        self.fix_spot_size = bool(owner.fix_spot_size.value)
        self.astigmatic = bool(owner.astigmatic.value)
        self.accuracy = owner.accuracy
        self.direction = owner.direction

    def __output(self, ws):
        E = np.asarray(ws.mtx.node_field_vector(ws.node_idx, ws.f_idx))
        # Directly accessing the node q doesn't work during
        # a simulation as they are not updated
        # qx = ws.oinfo.nodes[0].qx
        # qy = ws.oinfo.nodes[0].qy
        # Neither does accessing the last trace
        # qx = ws.sim.model.last_trace[ws.oinfo.nodes[0]].qx
        # qy = ws.sim.model.last_trace[ws.oinfo.nodes[0]].qy
        qx, qy = ws.sim.get_q(ws.oinfo.nodes[0])
        try:
            result = np.asarray(
                optimise_HG00_q_scipy(
                    E,
                    (qx, qy),
                    ws.sim.model.homs,
                    fix_spot_size=ws.fix_spot_size,
                    astigmatic=ws.astigmatic,
                    accuracy=self.accuracy,
                    full_output=False,
                )
            )
        except ConvergenceException:
            q = BeamParam(w0=np.nan, z=np.nan)
            result = np.array([q, q])

        if ws.direction == "both":
            return result
        elif ws.direction == "x":
            return result[0]
        else:
            return result[1]


@float_parameter("f", "Frequency", units="Hz")
@bool_parameter(
    "fix_spot_size",
    "Fix spot size",
    units="",
)
@bool_parameter(
    "astigmatic",
    "Astigmatic",
    units="",
)
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class OptimalQ(Detector):
    """This detector tries to compute an optimal beam parameter (:math:`q`) for a
    specified optical frequency at a node.

    Output of this detector into an array solution will be a tuple
    of :class:`.BeamParam` in each transverse direction, (qx, qy).
    If the optimisation process fails beam parameter objects will
    NaN values will be returned.

    Parameters
    ----------
    name : str
        Name of the detector
    node : [str | finesse.components.node]
        Node name or object to put this detector at
    f : float
        Frequency component tro compute the optimal beam parameter for.
    fix_spot_size : bool, optional
        When True the optimised will keep the current spot size at the node
        fixed and just optimise the curvature.
    astigmatic : bool, optional
        When True qx and qy will be optimised separately
    accuracy : float, optional
        Approximate mismatch accuracy to try and compute the optimised beam
        parameter to. mismatch(q_actual, q_optimal) < accuracy
    direction : str, optional
        Return either `both` or just the `x` or `y` modes

    Notes
    -----
    This method uses the :func:`finesse.gaussian.optimise_HG00_q`
    or :func:`finesse.gaussian.optimise_HG00_q_scipy` for optimising the HG mode amplitudes at
    the node and frequency requested. This particular method finds a new
    set of {qx, qy} values which maximise the HG00 mode content, whilst
    reducing the HG20 and HG02 mode content.

    ``Failure``

    If the optical field being optimised does not have a HG00 like
    appearance then. For example, trying to optimise the shape of
    an RF sideband field inside a cavity that it is not resonant in.
    """

    def __init__(
        self,
        name,
        node: Node,
        f,
        *,
        fix_spot_size=False,
        astigmatic=False,
        accuracy=1e-9,
        direction="both",
    ):
        if node.type is not NodeType.OPTICAL:
            raise Exception(f"Must be an optical node used for OptimalQ {name}")
        Detector.__init__(
            self, name, node, shape=(2,), dtype=BeamParam, label="Optimal q"
        )
        self.f = f
        self.fix_spot_size = fix_spot_size
        self.astigmatic = astigmatic
        self.accuracy = accuracy
        self.direction = direction

    @property
    def direction(self):
        """Sets the output of the detector.

        If `both` then both qx and qy are returned. if `x` or `y` are used then only the
        required one is returned.
        """
        return self.__direction

    @direction.setter
    def direction(self, value):
        if value not in ("both", "x", "y"):
            raise FinesseException("Direction should be `both`, `x`, or `y`")
        self.__direction = value
        if value == "both":
            self._update_dtype_shape((2,))
        else:
            self._update_dtype_shape((1,))

    def _get_workspace(self, sim):
        if not sim.is_modal:
            raise FinesseException(
                f"OptimalQ detector {self} needs higher order modes to be enabled."
            )
        if (2, 0) not in sim.model.mode_index_map:
            raise FinesseException(
                f"OptimalQ detector {self} needs HG20 mode in the simulation"
            )
        if (0, 2) not in sim.model.mode_index_map:
            raise FinesseException(
                f"OptimalQ detector {self} needs HG02 mode in the simulation"
            )
        return OptimalQWorkspace(self, sim)



cimport cython

import logging

import numpy as np
cimport numpy as np

from ...env import warn

from finesse.cymath.complex cimport carg, cexp
from scipy.linalg.cython_blas cimport zscal, zgemm

from libc.stdlib cimport free, calloc
from cpython.ref cimport PyObject

from finesse.knm cimport (
    KnmMatrix,
    knm_bh_workspace,
    knm_bh_ws_init,
    knm_bh_ws_free,
    knm_bh_ws_is_changing,
    knm_bh_ws_recompute_mismatch,
    knm_bh_ws_recompute,
    knm_bh_ws_recompute_misalignment,
)
from finesse.knm.bayerhelms cimport fast_compute_knm_matrix_bh
from finesse.cymath.gaussbeam cimport bp_beamsize
from finesse.knm.matrix cimport knm_loss, c_flip_odd_horizontal
from finesse.cymath.cmatrix cimport SubCCSView
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper, NoiseInfo
from finesse.components.general import NodeType, NoiseType
from finesse.cymath.gaussbeam cimport c_transform_q
from finesse.cymath.homs cimport unm_ws_recache_from_bp, unm_factor_store_recache
from finesse.frequency cimport frequency_info_t
from finesse.simulations.base cimport NodeBeamParam
from finesse.knm.maps cimport (
    set_knm_map_workspace,
    free_knm_map_workspace,
    init_knm_map_workspace,
    c_scattering_coefficients_to_KnmMatrix,
    update_map_data_in_workspace
)
from finesse.knm.integrators cimport c_riemann_optimised, c_outer_conj_product_2, update_U_xy_array
from finesse.utilities.collections cimport OrderedSet


LOGGER = logging.getLogger(__name__)

cdef double[:,::1] abcd_unity = np.eye(2, dtype=float)


cdef class KnmConnectorWorkspace(ConnectorWorkspace):
    def __cinit__(self, *args, **kwargs):
        self.onode_ids = NULL
        self.oconn_info = NULL
        self.Kws = NULL

    def __init__(self, object owner, object sim, *args, **kwargs):
        from finesse.simulations.simulation import BaseSimulation
        assert(isinstance(sim, BaseSimulation))
        ConnectorWorkspace.__init__(self, owner, sim, *args, **kwargs)

        # Here we automatically generate the coupling strings used
        # for generating the Knm matricies. We use the optical to optical
        # connections that have been registered, the port number is taken
        # from the order in which they are added to the component. I guess
        # there could also be a more explicit defition of the port "index"
        # given in the _add_port method but we'll use this for now.

        # TODO ddb could store index in port object perhaps, along with the
        # list of optical ports
        self.o2o = owner.all_internal_optical_connections
        oports = list(p for p in owner.ports if p.type == NodeType.OPTICAL)

        self.N_opt_ports = len(oports)
        self.N_opt_conns = len(self.o2o)

        if self.onode_ids != NULL or self.oconn_info != NULL or self.Kws != NULL:
            raise MemoryError()

        self.onode_ids = <Py_ssize_t*> calloc(self.N_opt_ports * 2, sizeof(Py_ssize_t))
        if not self.onode_ids:
            raise MemoryError()

        self.oconn_info = <KnmInfo*> calloc(self.N_opt_conns, sizeof(KnmInfo))
        if not self.oconn_info:
            raise MemoryError()

        # 2 lots per connection because of x and y planes
        self.Kws = <knm_bh_workspace*> calloc(2*self.N_opt_conns, sizeof(knm_bh_workspace))
        if not self.Kws:
            raise MemoryError()

        # Indexes for access the simulation trace, 2 because we have inputs and outputs
        self.trace_node_indices = <Py_ssize_t*> calloc(self.N_opt_ports * 2, sizeof(Py_ssize_t))
        if not self.trace_node_indices:
            raise MemoryError()

        # need these node ids for the simulation for indexing traces
        # Signal and carrier all have the same optical node IDs so
        # use carrier here because it should always exist
        for i, p in enumerate(oports):
            self.onode_ids[2*i] = sim.carrier.node_id(p.i)
            self.onode_ids[2*i+1] = sim.carrier.node_id(p.o)
            self.trace_node_indices[2*i] = sim.trace_node_index[p.i]
            self.trace_node_indices[2*i+1] = sim.trace_node_index[p.o]

        cdef np.ndarray[double, ndim=1, mode="c"] scatter_loss
        self.__knm_matrices = []

        for i, (conn, (f, t)) in enumerate(self.o2o.items()):
            # TODO ddb should probably use some fixed index rather than a list of the order
            # they are defined in the element definition
            a = oports.index(f.port)
            b = oports.index(t.port)
            coupling = f"{a+1}{b+1}"
            self.oconn_info[i].ptr_owner = <PyObject*>self
            self.oconn_info[i].from_port_idx = a
            self.oconn_info[i].to_port_idx = b
            self.oconn_info[i].K_ws_x = &self.Kws[2*i]
            self.oconn_info[i].K_ws_y = &self.Kws[2*i+1]
            self.oconn_info[i].abcd_x = &abcd_unity[0,0]
            self.oconn_info[i].abcd_y = &abcd_unity[0,0]
            # The final matrix to be used for filling
            knm = KnmMatrix(self.sim.model_settings.homs_view, self.owner.name, coupling)
            self.oconn_info[i].knm_mtx = &knm.mtx
            self.__knm_matrices.append(knm) # store these objects for access later if needed
            try:
                setattr(self, f"K{coupling}", knm)
            except AttributeError:
                raise AttributeError(f"Cannot set K{coupling} in {self}. Check that this workspace is a Python object or if it is a Cython workspace that K{coupling} is defined.")

            # Make the buffer of knm loss data
            scatter_loss = np.zeros(self.sim.model_settings.num_HOMs)
            try:
                setattr(self, f"K{coupling}_loss", scatter_loss)
            except AttributeError:
                raise AttributeError(f"Cannot set K{coupling}_loss in {self}. Check that this workspace is a Python object or if it is a Cython workspace that K{coupling}_loss is defined.")
            # Keep ptr to the buffer for use in compute_knm_losses
            self.oconn_info[i].loss = &scatter_loss[0]

        self.total_losses = np.zeros(self.sim.model_settings.num_HOMs)
        if sim.signal:
            self.signal.set_fill_noise_function(NoiseType.QUANTUM, optical_quantum_noise_knm)


    def __dealloc__(self):
        cdef:
            int i

        if self.onode_ids:
            free(self.onode_ids)

        if self.oconn_info:
            for i in range(self.N_opt_conns):
                if self.oconn_info[i].use_map:
                    free_knm_map_workspace(&self.oconn_info[i].map_ws)
            free(self.oconn_info)

        free(self.trace_node_indices)

        if self.Kws:
            for i in range(2*self.N_opt_conns):
                knm_bh_ws_free(&self.Kws[i])

            free(self.Kws)

        if self.__knm_matrices:
            self.__knm_matrices.clear()


    cpdef set_knm_info(self, connection,
        double[:,::1] abcd_x = abcd_unity,
        double[:,::1] abcd_y = abcd_unity,
        double nr_from=1, double nr_to=1,
        bint is_transmission=False,
        Parameter beta_x = None,
        double beta_x_factor = 0,
        Parameter beta_y = None,
        double beta_y_factor = 0,
        Parameter alpha = None,
        Map apply_map = None,
        double map_phase_factor = 1,
        bint map_fliplr = False
    ) :
        """
        Python facing method to set how Knm calculations should be handled for a
        particular connection. When the workspace is being created this should be called
        if modal simualtions are being used. This then sets which ABCD matricies to use,
        various parameters required, and whether the connection is a transmission or
        reflection.

        This method should only be called once with the required settings. Calling it
        again will rewrite settings for this connection.

        Parameters
        ----------
        abcd_x, abcd_y : double[:,::1]
            ABCD matrices that transform some input beam parameter to the output
            of this connection

        nr_from, nr_to : double
            refractive index at the input and output of this connection

        is_transmission : bool
            True if this connection represents a transmission through an element

        beta_x, beta_y : Parameter
            Parameter objects that reference a yaw (beta_x) and pitch (beta_y)
            misalignment

        beta_x_factor, beta_y_factor : double
            Scaling factor for misalignment parameters

        alpha : Parameter
            Macroscopic angle of incidence parameter for the element

        apply_map : Map
            When set this will apply the map object to this connection

        map_fliplr : bool
            When True the map is flipped left-to-right when computing the scattering
            coefficients
        """
        cdef:
            KnmInfo *conn
            int index
            object model = self.sim.model

        if abcd_x.shape[0] != abcd_x.shape[1] != 2:
            raise Exception("ABCD X is not 2x2")
        if abcd_y.shape[0] != abcd_y.shape[1] != 2:
            raise Exception("ABCD Y is not 2x2")
        # First find the connection str in our connections then store the information)
        if connection not in self.o2o:
            raise Exception(f"Connection {connection} is not a valid connection in the element {self.owner}")

        index = list(self.o2o.keys()).index(connection)
        conn = &self.oconn_info[index]
        conn.has_been_set = True
        conn.abcd_x = &abcd_x[0, 0]
        conn.abcd_y = &abcd_y[0, 0]
        conn.is_transmission = is_transmission
        conn.nr_from = nr_from
        conn.nr_to = nr_to
        conn.use_map = apply_map is not None
        conn.use_bh = True
        conn.bhelms_mtx_func = fast_compute_knm_matrix_bh

        if conn.use_map:
            set_knm_map_workspace(model, &conn.map_ws, apply_map, self.sim.model_settings.k0, map_phase_factor)
            conn.map_fliplr = map_fliplr

        if beta_x is not None:
            # If the value of the parameter is a symbolic reference, at this point
            # the c_value of the parameter will not be updated yet
            # https://gitlab.com/ifosim/finesse/finesse3/-/issues/691
            beta_x._reset_cvalue()
            conn.beta_x = &beta_x.__cvalue
            conn.beta_x_is_changing = beta_x.is_changing if beta_x is not None else False
            conn.beta_x_factor = beta_x_factor
        if beta_y is not None:
            # If the value of the parameter is a symbolic reference, at this point
            # the c_value of the parameter will not be updated yet
            # https://gitlab.com/ifosim/finesse/finesse3/-/issues/691
            beta_y._reset_cvalue()
            conn.beta_y = &beta_y.__cvalue
            conn.beta_y_is_changing = beta_y.is_changing if beta_y is not None else False
            conn.beta_y_factor = beta_y_factor
        if alpha is not None:
            # If the value of the parameter is a symbolic reference, at this point
            # the c_value of the parameter will not be updated yet
            # https://gitlab.com/ifosim/finesse/finesse3/-/issues/691
            alpha._reset_cvalue()
            conn.alpha = &alpha.__cvalue
            conn.alpha_is_changing = alpha.is_changing
            if alpha.is_changing:
                raise NotImplementedError("Changing angle of incidence not supported yet")


    cdef initialise_knm_workspaces(KnmConnectorWorkspace self) :
        cdef:
            NodeBeamParam *q_from
            NodeBeamParam *q_to
            NodeBeamParam *q1
            NodeBeamParam *q2
            KnmInfo *info
            KnmMatrix knm
            # From Beam parameters transformed by ABCD
            complex_t qx_from_trns, qy_from_trns

            double lambda0 = self.sim.model_settings.lambda0
            int maxtem = max(self.sim.model_settings.max_n, self.sim.model_settings.max_m)
            int i = 0
            double beta_x, beta_y

        oconn_names = tuple(self.o2o.keys())
        # Loop over the node ids stored in input, output pairs and get the
        # beam param data
        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            coupling = f"{info.from_port_idx+1}{info.to_port_idx+1}"

            if not info.has_been_set:
                raise Exception(f"Information for connection {oconn_names[i]} at {self.owner.name} has not been set with `set_knm_info`")

            if info.abcd_x == NULL:
                raise Exception(f"Knm info ABCDx for connection {oconn_names[i]} at {self.owner.name} is NULL")
            if info.abcd_y == NULL:
                raise Exception(f"Knm info ABCDy for connection {oconn_names[i]} at {self.owner.name} is NULL")
            if info.beta_x == NULL and info.beta_x_factor != 0:
                raise Exception(f"Knm info beta_x for connection {oconn_names[i]} at {self.owner.name} is NULL but has a beta_x_factor set")
            if info.beta_y == NULL and info.beta_y_factor != 0:
                raise Exception(f"Knm info beta_y for connection {oconn_names[i]} at {self.owner.name} is NULL but has a beta_y_factor set")

            q_from = &self.sim.trace[self.trace_node_indices[2*info.from_port_idx]]
            q_to = &self.sim.trace[self.trace_node_indices[2*info.to_port_idx+1]]
            # Get any misalignment factor is one has been set
            beta_x = info.beta_x[0] if info.beta_x else 0
            beta_y = info.beta_y[0] if info.beta_y else 0

            info.use_bh = True # always use/initialise BH for now

            if not info.use_map or (info.use_map and not info.map_ws.is_focusing_element):
                # we're using a map and it's not a focusing element map, or just not using a map
                # then use BH for mode projection.
                # 'From' Beam parameters after propagation through the connections ABCD
                qx_from_trns = c_transform_q(info.abcd_x, q_from.qx.q, info.nr_from, info.nr_to)
                qy_from_trns = c_transform_q(info.abcd_y, q_from.qy.q, info.nr_from, info.nr_to)
            elif info.use_map and info.map_ws.is_focusing_element:
                # BH is just doing tilts so use output mode
                qx_from_trns = q_to.qx.q
                qy_from_trns = q_to.qy.q

            # print("")
            # print("! Knm flags", self.owner.name)
            # print("! Connection port", info.from_port_idx+1, "->", info.to_port_idx+1)
            # print("! use BH", info.use_bh)
            # print("! use map", info.use_map)
            # print("! q_from changing", q_from.is_changing)
            # print("! q_to changing", q_to.is_changing)
            # print("! beta_x changing", info.beta_x_is_changing)
            # print("! beta_y changing", info.beta_y_is_changing)
            # print("! map changing", info.map_ws.map_is_changing)

            # Initialise the Knm workspaces with the current values
            if info.use_bh:
                knm_bh_ws_init(
                    info.K_ws_x, qx_from_trns, q_to.qx.q, beta_x, info.beta_x_factor, info.nr_to, lambda0, maxtem
                )
                knm_bh_ws_init(
                    info.K_ws_y, qy_from_trns, q_to.qy.q, beta_y, info.beta_y_factor, info.nr_to, lambda0, maxtem,
                )

            if info.use_map:
                # if the map is marked as a focusing element then
                # it does the from-to mode projection, otherwise BH
                # does it
                if not info.map_ws.is_focusing_element:
                    q1 = q_from
                    q2 = q_from
                else:
                    q1 = q_from
                    q2 = q_to

                init_knm_map_workspace(
                    &info.map_ws,
                    max(self.sim.model_settings.max_m, self.sim.model_settings.max_n)+1,
                    q1, q2,
                    # BH already includes the reflection minus flip
                    not info.is_transmission and not info.use_bh
                )

            if info.use_bh and not info.use_map:
                knm = getattr(self, f"K{coupling}")
                info.bh_mtx = &knm.mtx
                info.map_mtx = NULL
            elif not info.use_bh and info.use_map:
                knm = getattr(self, f"K{coupling}")
                info.bh_mtx = NULL
                info.map_mtx = &knm.mtx
            else:
                # if we have a map, then we will probably have both BH and map calcuations.
                # in this case we need to compute multiple scattering matrices then
                # multiply them together to get the matrix used for filling.
                knm = KnmMatrix(self.sim.model_settings.homs_view, self.owner.name, coupling + "_bh")
                info.bh_mtx = &knm.mtx
                self.__knm_matrices.append(knm) # keep a reference so doesn't get cleaned up
                # Same again for map matrix
                knm = KnmMatrix(self.sim.model_settings.homs_view, self.owner.name, coupling + "_map")
                info.map_mtx = &knm.mtx
                self.__knm_matrices.append(knm) # keep a reference so doesn't get cleaned up


    cpdef update_map_data(KnmConnectorWorkspace self):
        cdef:
            KnmInfo *info
        # not the most efficient way to loop over all possible
        # Knm connections and look for maps, but this doesn't
        # get called much...
        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            if info.use_map and info.map_ws.map_is_changing:
                update_map_data_in_workspace(&info.map_ws)


    cpdef flag_changing_beam_parameters(
        KnmConnectorWorkspace self,
        OrderedSet changing_mismatch_edges
    ):
        """Iterate through all optical connections and checks/flag which optical
        connections need to having their scattering matrices recalculated. This
        occurs if the beam tracer traces some new q-parameters, or if an ABCD has
        a changing parameter, or if some optical path distortion map is changing.

        Parameters
        ----------
        changing_couplings : tuple
            Tuple of (node_in, node_out) which must be recomputed.

        Notes
        -----
        `changing_couplings` typically comes from the trace forest`s method
        `trace_forest.find_potential_mismatch_couplings`. Which determines
        potential changes in mode scattering parameters, from changing ABCD
        properties.
        """
        cdef:
            KnmInfo *info
            bint is_mm_changing
            Py_ssize_t ni_idx, no_idx
            const NodeBeamParam* q_from
            const NodeBeamParam* q_to

        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            if info.use_bh:
                ni_idx = self.onode_ids[2*info.from_port_idx]
                no_idx = self.onode_ids[2*info.to_port_idx+1]
                q_from = &self.sim.trace[self.trace_node_indices[2*info.from_port_idx]]
                q_to = &self.sim.trace[self.trace_node_indices[2*info.to_port_idx+1]]

                is_mm_changing = (
                    not q_from.is_fixed or not q_to.is_fixed
                )
                # Need to determine if ABCD is changing here, as that
                # will also change how q_from is projected into q_to
                if (ni_idx, no_idx) in changing_mismatch_edges:
                    is_mm_changing = True

                info.K_ws_x.is_mm_changing = is_mm_changing
                info.K_ws_y.is_mm_changing = is_mm_changing
                info.K_ws_x.is_alignment_changing = info.beta_x_is_changing
                info.K_ws_y.is_alignment_changing = info.beta_y_is_changing

                if knm_bh_ws_is_changing(info.K_ws_x) or knm_bh_ws_is_changing(info.K_ws_y):
                    LOGGER.debug(f"{self.owner.name}.K{info.from_port_idx+1}{info.to_port_idx+1} is changing")

    cdef void update_changing_knm_workspaces(KnmConnectorWorkspace self) noexcept:
        cdef:
            NodeBeamParam *q_from
            NodeBeamParam *q_to
            KnmInfo *info
            complex_t qx_from_trns, qy_from_trns

        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            if info.use_bh:
                q_from = &self.sim.trace[self.trace_node_indices[2*info.from_port_idx]]
                q_to = &self.sim.trace[self.trace_node_indices[2*info.to_port_idx+1]]

                # if there's some changing Knm then we need to recompute it. Which parts need
                # to be recomputed are decided here then performed
                if info.K_ws_x.is_mm_changing:
                    # If we use a map which is doing the focussing, do the mode projection there
                    if info.use_map and info.map_ws.is_focusing_element:
                        qx_from_trns = q_to.qx.q
                    else:
                        qx_from_trns = c_transform_q(info.abcd_x, q_from.qx.q, info.nr_from, info.nr_to)

                    if info.beta_x == NULL:
                        knm_bh_ws_recompute_mismatch(info.K_ws_x, qx_from_trns, q_to.qx.q)
                    else:
                        knm_bh_ws_recompute(info.K_ws_x, qx_from_trns, q_to.qx.q, info.beta_x[0])

                elif info.K_ws_x.is_alignment_changing and info.beta_x != NULL:
                    knm_bh_ws_recompute_misalignment(info.K_ws_x, info.beta_x[0])

                if info.K_ws_y.is_mm_changing:
                    # If we use a map which is doing the focussing, do the mode projection there
                    if info.use_map and info.map_ws.is_focusing_element:
                        qy_from_trns = q_to.qy.q
                    else:
                        qy_from_trns = c_transform_q(info.abcd_y, q_from.qy.q, info.nr_from, info.nr_to)

                    if info.beta_y == NULL:
                        knm_bh_ws_recompute_mismatch(info.K_ws_y, qy_from_trns, q_to.qy.q)
                    else:
                        knm_bh_ws_recompute(info.K_ws_y, qy_from_trns, q_to.qy.q, info.beta_y[0])

                elif info.K_ws_y.is_alignment_changing and info.beta_y != NULL:
                    knm_bh_ws_recompute_misalignment(info.K_ws_y, info.beta_y[0])


    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void compute_scattering_matrices(KnmConnectorWorkspace self) noexcept:
        cdef:
            KnmInfo *info = NULL
            Py_ssize_t i = 0
            knm_map_workspace *map_ws = NULL
            bint map_q_changed = False
            complex_t alpha, beta
            int N, inc
            bint updated
            bint is_reflection

        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            updated = False # whether the knm has been updated or not
            is_reflection = not info.is_transmission

            if info.use_bh and (knm_bh_ws_is_changing(info.K_ws_x) or knm_bh_ws_is_changing(info.K_ws_y)):
                info.bhelms_mtx_func(
                    info.K_ws_x,
                    info.K_ws_y,
                    &(self.sim.model_settings.homs_view[0,0]),
                    info.bh_mtx.ptr,
                    self.sim.model_settings.num_HOMs,
                    self.sim.config_data.nthreads_homs,
                )
                updated = True

            if info.use_map:
                map_ws = &info.map_ws
                map_q_changed = False
                if map_ws.new_map_data or not map_ws.q_from.is_fixed or not map_ws.q_to.is_fixed:
                    if not map_ws.q_from.is_fixed:
                        unm_ws_recache_from_bp(map_ws.uiws, &map_ws.q_from.qx, &map_ws.q_from.qy)
                        unm_factor_store_recache(
                            map_ws.unm_i_factor_ws, map_ws.uiws,
                            True, # remove Gouy
                            # we don't actually flip the map here, we just
                            # flip the modes relative to the map
                            info.map_fliplr,
                        )
                    if not map_ws.q_to.is_fixed:
                        unm_ws_recache_from_bp(map_ws.uows, &map_ws.q_to.qx, &map_ws.q_to.qy)
                        unm_factor_store_recache(
                            map_ws.unm_o_factor_ws,
                            map_ws.uows,
                            True,
                            info.map_fliplr
                        )

                    w = min(
                        bp_beamsize(&map_ws.q_to.qx),
                        bp_beamsize(&map_ws.q_to.qy),
                        bp_beamsize(&map_ws.q_from.qx),
                        bp_beamsize(&map_ws.q_from.qy),
                    )

                    if w/map_ws.dx < 10 or w/map_ws.dy < 10:
                        warn(
                            "Spot size vs map resolution is low, increase map resolution or spot size\n" +
                            f"for {self.owner.name} connection {info.from_port_idx+1}->{info.to_port_idx+1}"
                        )

                    # Then update the Un and Um arrays
                    update_U_xy_array(
                        map_ws.x, map_ws.Nx,
                        map_ws.y, map_ws.Ny,
                        map_ws.Un, map_ws.Um, map_ws.Nm,
                        map_ws.uiws, map_ws.unm_i_factor_ws
                    )
                    update_U_xy_array(
                        map_ws.x, map_ws.Nx,
                        map_ws.y, map_ws.Ny,
                        map_ws.Un_, map_ws.Um_, map_ws.Nm,
                        map_ws.uows, map_ws.unm_o_factor_ws
                    )
                    # update the outer products for all modes for Un and Um
                    c_outer_conj_product_2(map_ws.Nm, map_ws.Ny, map_ws.Um, map_ws.Um_, map_ws.Umm_)
                    c_outer_conj_product_2(map_ws.Nm, map_ws.Nx, map_ws.Un, map_ws.Un_, map_ws.Unn_)
                    map_q_changed = True

                if map_ws.new_map_data or map_q_changed:
                    map_ws.new_map_data = False # mark that we've calculated this at least once
                    updated = True
                    # calculate new scattering coefficients using the optimised
                    # method using BLAS functions
                    c_riemann_optimised(
                        map_ws.Nx, map_ws.Ny, map_ws.Nm,
                        map_ws.dx * map_ws.dy,
                        map_ws.z,
                        map_ws.Unn_,
                        map_ws.Umm_,
                        map_ws.tmp,
                        map_ws.K,
                    )
                    # convert the 4D scattering coefficients into the modal matrix for filling
                    c_scattering_coefficients_to_KnmMatrix(
                        self.sim.model_settings.homs_view,
                        map_ws.Nm,
                        map_ws.K,
                        info.map_mtx
                    )

            # only do these steps if a knm matrix has been updated. Currently
            # this loop is called constantly for every element, regardless of
            # whether it is changing or not
            if updated:
                if info.use_bh and info.use_map:
                    # using two Knm so need to multiply them
                    alpha = 1
                    beta = 0 # rewrite output matrix
                    N = info.bh_mtx.size1
                    zgemm(
                        "T", "T", # KnmMatrices are C ordered, but zgemm expects F ordered
                        &N, &N, &N, # Square matrix
                        &alpha, # include any phase removal
                        info.bh_mtx.ptr, &N,
                        info.map_mtx.ptr, &N,
                        &beta, info.knm_mtx.ptr, &N
                    )

                if self.sim.model_settings.phase_config.zero_k00:
                    # final step zero 00->00 phase, which is just removing a plane
                    # wave phase from the entire scattering calculation
                    #  - arg of 00->00 term
                    #  - create unit complex vector with -1 * arg
                    #  - apply to all elements in matrix
                    alpha = cexp(-1j*carg(info.knm_mtx.ptr[0]))
                    N = info.knm_mtx.size1 * info.knm_mtx.size2
                    inc = 1 # DenseZMatrix is memory contiguous
                    zscal(&N, &alpha, info.knm_mtx.ptr, &inc)

                if is_reflection:
                    # if we're doing a reflection we need to apply the
                    # parity operator
                    c_flip_odd_horizontal(
                        info.knm_mtx,
                        self.sim.model_settings.homs_view
                    )

                # Note: no need for reversing gouy here compared to
                # Finesse v2. The BH and map calculations do that
                # themselves - for the map it's essentially free
                # as it's just not added in the first place


    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void compute_knm_losses(KnmConnectorWorkspace self) noexcept nogil:
        cdef:
            KnmInfo *info
            Py_ssize_t i

        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            #if knm_bh_ws_is_changing(info.K_ws_x) or knm_bh_ws_is_changing(info.K_ws_y):
            # use full knm matrix to compute the HOM losses, could add some way to
            # compute if this has changed
            knm_loss(info.knm_mtx.ptr, info.loss, self.sim.model_settings.num_HOMs)


# N.B. This function depends on the order of ports / nodes being consistent between
# ws.onode_ids and ws.oconn_info, and may break if this order is changed.
# TODO: This doesn't consider the couplings in the signal matrix, effectively assuming they're 1,
# which isn't always correct. In such cases a custom function will be needed such
# as for a mirror or beamsplitter.
optical_quantum_noise_knm = FillFuncWrapper.make_from_ptr(c_optical_quantum_noise_knm)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef object c_optical_quantum_noise_knm(ConnectorWorkspace cws) :
    cdef:
        KnmConnectorWorkspace ws = <KnmConnectorWorkspace> cws
        NoiseInfo noises = ws.output_noise
        frequency_info_t *freq

        Py_ssize_t i, j, k, h
        Py_ssize_t port_idx

        complex_t factor

    if ws.sim.signal.nhoms == 0:
        return
    for i in range(ws.sim.signal.optical_frequencies.size):
        freq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        factor = 0.5 * (1 + freq.f_car[0] / ws.sim.model_settings.f0)
        for j in range(noises.num_nodes):
            port_idx = -1
            for k in range(ws.N_opt_ports):
                if ws.onode_ids[2 * k + 1] == noises.node_info[j].idx:
                    port_idx = k
                    break
            if port_idx == -1:
                continue
            ws.total_losses[:] = 0
            for k in range(ws.N_opt_conns):
                if ws.oconn_info[k].to_port_idx == port_idx:
                    for h in range(ws.sim.signal.nhoms):
                        ws.total_losses[h] += ws.oconn_info[k].loss[h]
            (<SubCCSView>noises.ptrs[j][freq.index]).fill_za_dv(factor, ws.total_losses)



import pytest


# Override default element to take two arguments.
@pytest.fixture
def fake_element_cls(fake_element_cls):
    class FakeElement(fake_element_cls):
        def __init__(self, name, a=None, b=None):
            super().__init__(name)
            self.a = a
            self.b = b

    return FakeElement


@pytest.fixture
def fake_variadic_element_cls(fake_element_cls):
    class FakeVariadicElement(fake_element_cls):
        def __init__(self, name, a, *args, b=None, **kwargs):
            super().__init__(name)
            self.a = a
            self.args = args
            self.b = b
            self.kwargs = kwargs

    return FakeVariadicElement


@pytest.fixture
def spec(
    spec, fake_element_adapter_factory, fake_element_cls, fake_variadic_element_cls
):
    spec.register_element(
        fake_element_adapter_factory(
            fake_element_cls, full_name="fake_element", short_name="fake"
        )
    )
    spec.register_element(
        fake_element_adapter_factory(
            fake_variadic_element_cls,
            full_name="fake_variadic_element",
            short_name="fake_var",
        )
    )
    return spec


@pytest.mark.parametrize(
    "name,args,kwargs,argument_defaults,prefer_keywords,expected",
    (
        # No optional arguments.
        ("myel", [], {}, False, False, "fake_element myel"),
        ("myel", [], {}, False, True, "fake_element myel"),
        ("myel", [], {}, True, False, "fake_element myel none none"),
        ("myel", [], {}, True, True, "fake_element myel a=none b=none"),
        # Optional arguments as positional.
        ("myel", [None], {}, False, False, "fake_element myel"),
        ("myel", [None], {}, False, True, "fake_element myel"),
        ("myel", [None], {}, True, False, "fake_element myel none none"),
        ("myel", [None], {}, True, True, "fake_element myel a=none b=none"),
        ("myel", [None, None], {}, False, False, "fake_element myel"),
        ("myel", [None, None], {}, False, True, "fake_element myel"),
        ("myel", [None, None], {}, True, False, "fake_element myel none none"),
        ("myel", [None, None], {}, True, True, "fake_element myel a=none b=none"),
        # Mixed form optional arguments.
        ("myel", [None], {"b": None}, False, False, "fake_element myel"),
        ("myel", [None], {"b": None}, False, True, "fake_element myel"),
        ("myel", [None], {"b": None}, True, False, "fake_element myel none none"),
        ("myel", [None], {"b": None}, True, True, "fake_element myel a=none b=none"),
        ("myel", [], {"a": None}, False, False, "fake_element myel"),
        ("myel", [], {"a": None}, False, True, "fake_element myel"),
        ("myel", [], {"a": None}, True, False, "fake_element myel none none"),
        ("myel", [], {"a": None}, True, True, "fake_element myel a=none b=none"),
        ("myel", [], {"b": None}, False, False, "fake_element myel"),
        ("myel", [], {"b": None}, False, True, "fake_element myel"),
        ("myel", [], {"b": None}, True, False, "fake_element myel none none"),
        ("myel", [], {"b": None}, True, True, "fake_element myel a=none b=none"),
        ("myel", [], {"a": None, "b": None}, False, False, "fake_element myel"),
        ("myel", [], {"a": None, "b": None}, False, True, "fake_element myel"),
        (
            "myel",
            [],
            {"a": None, "b": None},
            True,
            False,
            "fake_element myel none none",
        ),
        (
            "myel",
            [],
            {"a": None, "b": None},
            True,
            True,
            "fake_element myel a=none b=none",
        ),
        ("myel", [1], {}, False, False, "fake_element myel 1"),
        ("myel", [1], {}, False, True, "fake_element myel a=1"),
        ("myel", [1], {}, True, False, "fake_element myel 1 none"),
        ("myel", [1], {}, True, True, "fake_element myel a=1 b=none"),
        ("myel", [1, 2], {}, False, False, "fake_element myel 1 2"),
        ("myel", [1, 2], {}, False, True, "fake_element myel a=1 b=2"),
        ("myel", [1, 2], {}, True, False, "fake_element myel 1 2"),
        ("myel", [1, 2], {}, True, True, "fake_element myel a=1 b=2"),
        # Optional arguments as keyword.
        ("myel", [], {"a": 1}, False, False, "fake_element myel 1"),
        ("myel", [], {"a": 1}, False, True, "fake_element myel a=1"),
        ("myel", [], {"a": 1}, True, False, "fake_element myel 1 none"),
        ("myel", [], {"a": 1}, True, True, "fake_element myel a=1 b=none"),
        ("myel", [], {"a": 1, "b": 2}, False, False, "fake_element myel 1 2"),
        (
            "myel",
            [],
            {"a": 1, "b": 2},
            False,
            True,
            "fake_element myel a=1 b=2",
        ),
        ("myel", [], {"a": 1, "b": 2}, True, False, "fake_element myel 1 2"),
        ("myel", [], {"a": 1, "b": 2}, True, True, "fake_element myel a=1 b=2"),
    ),
)
def test_element(
    unbuilder,
    model,
    element_dump,
    fake_element_cls,
    name,
    args,
    kwargs,
    argument_defaults,
    prefer_keywords,
    expected,
):
    """Element with normal signature."""
    model.add(fake_element_cls(name, *args, **kwargs))
    dump = next(iter(element_dump("fake_element", fake_element_cls, model)))
    script = unbuilder.unbuild(
        dump,
        argument_defaults=argument_defaults,
        prefer_keywords=prefer_keywords,
    )
    assert script == expected


@pytest.mark.parametrize(
    "name,args,kwargs,argument_defaults,prefer_keywords,expected",
    (
        # No optional arguments.
        ("myel", [None], {}, False, False, "fake_variadic_element myel none"),
        ("myel", [None], {}, False, True, "fake_variadic_element myel a=none"),
        ("myel", [None], {}, True, False, "fake_variadic_element myel none b=none"),
        ("myel", [None], {}, True, True, "fake_variadic_element myel a=none b=none"),
        # Variadic positional arguments.
        (
            "myel",
            [None, None],
            {},
            False,
            False,
            "fake_variadic_element myel none none",
        ),
        ("myel", [None, None], {}, False, True, "fake_variadic_element myel none none"),
        (
            "myel",
            [None, None],
            {},
            True,
            False,
            "fake_variadic_element myel none none b=none",
        ),
        (
            "myel",
            [None, None],
            {},
            True,
            True,
            "fake_variadic_element myel none none b=none",
        ),
        # Mixed form optional arguments.
        ("myel", [None], {"b": None}, False, False, "fake_variadic_element myel none"),
        ("myel", [None], {"b": None}, False, True, "fake_variadic_element myel a=none"),
        (
            "myel",
            [None],
            {"b": None},
            True,
            False,
            "fake_variadic_element myel none b=none",
        ),
        (
            "myel",
            [None],
            {"b": None},
            True,
            True,
            "fake_variadic_element myel a=none b=none",
        ),
        ("myel", [], {"a": None}, False, False, "fake_variadic_element myel none"),
        ("myel", [], {"a": None}, False, True, "fake_variadic_element myel a=none"),
        (
            "myel",
            [],
            {"a": None},
            True,
            False,
            "fake_variadic_element myel none b=none",
        ),
        (
            "myel",
            [],
            {"a": None},
            True,
            True,
            "fake_variadic_element myel a=none b=none",
        ),
        (
            "myel",
            [],
            {"a": None, "b": None},
            False,
            False,
            "fake_variadic_element myel none",
        ),
        (
            "myel",
            [],
            {"a": None, "b": None},
            False,
            True,
            "fake_variadic_element myel a=none",
        ),
        (
            "myel",
            [],
            {"a": None, "b": None},
            True,
            False,
            "fake_variadic_element myel none b=none",
        ),
        (
            "myel",
            [],
            {"a": None, "b": None},
            True,
            True,
            "fake_variadic_element myel a=none b=none",
        ),
        (
            "myel",
            [],
            {"a": None, "b": None},
            False,
            False,
            "fake_variadic_element myel none",
        ),
        (
            "myel",
            [],
            {"a": None, "b": None},
            False,
            True,
            "fake_variadic_element myel a=none",
        ),
        ("myel", [1], {}, False, False, "fake_variadic_element myel 1"),
        ("myel", [1], {}, False, True, "fake_variadic_element myel a=1"),
        ("myel", [1], {}, True, False, "fake_variadic_element myel 1 b=none"),
        ("myel", [1], {}, True, True, "fake_variadic_element myel a=1 b=none"),
        ("myel", [1, 2], {}, False, False, "fake_variadic_element myel 1 2"),
        ("myel", [1, 2], {}, False, True, "fake_variadic_element myel 1 2"),
        ("myel", [1, 2], {}, True, False, "fake_variadic_element myel 1 2 b=none"),
        ("myel", [1, 2], {}, True, True, "fake_variadic_element myel 1 2 b=none"),
        # Optional arguments as keyword.
        ("myel", [], {"a": 1}, False, False, "fake_variadic_element myel 1"),
        ("myel", [], {"a": 1}, False, True, "fake_variadic_element myel a=1"),
        ("myel", [], {"a": 1}, True, False, "fake_variadic_element myel 1 b=none"),
        ("myel", [], {"a": 1}, True, True, "fake_variadic_element myel a=1 b=none"),
        (
            "myel",
            [],
            {"a": 1, "b": 2},
            False,
            False,
            "fake_variadic_element myel 1 b=2",
        ),
        (
            "myel",
            [],
            {"a": 1, "b": 2},
            False,
            True,
            "fake_variadic_element myel a=1 b=2",
        ),
        (
            "myel",
            [],
            {"a": 1, "b": 2},
            True,
            False,
            "fake_variadic_element myel 1 b=2",
        ),
        (
            "myel",
            [],
            {"a": 1, "b": 2},
            True,
            True,
            "fake_variadic_element myel a=1 b=2",
        ),
        # Variadic keyword arguments.
        (
            "myel",
            [],
            {"a": 1, "b": 2, "c": 3},
            False,
            False,
            "fake_variadic_element myel 1 b=2 c=3",
        ),
        (
            "myel",
            [],
            {"a": 1, "b": 2, "c": 3},
            False,
            True,
            "fake_variadic_element myel a=1 b=2 c=3",
        ),
        (
            "myel",
            [],
            {"a": 1, "b": 2, "c": 3},
            True,
            False,
            "fake_variadic_element myel 1 b=2 c=3",
        ),
        (
            "myel",
            [],
            {"a": 1, "b": 2, "c": 3},
            True,
            True,
            "fake_variadic_element myel a=1 b=2 c=3",
        ),
        # Variadic positional and keyword arguments.
        (
            "myel",
            [1, 2],
            {"b": 3, "c": 4},
            False,
            False,
            "fake_variadic_element myel 1 2 b=3 c=4",
        ),
        (
            "myel",
            [1, 2],
            {"b": 3, "c": 4},
            False,
            True,
            "fake_variadic_element myel 1 2 b=3 c=4",
        ),
        (
            "myel",
            [1, 2],
            {"b": 3, "c": 4},
            True,
            False,
            "fake_variadic_element myel 1 2 b=3 c=4",
        ),
        (
            "myel",
            [1, 2],
            {"b": 3, "c": 4},
            True,
            True,
            "fake_variadic_element myel 1 2 b=3 c=4",
        ),
    ),
)
def test_variadic_element(
    unbuilder,
    model,
    element_dump,
    fake_variadic_element_cls,
    name,
    args,
    kwargs,
    argument_defaults,
    prefer_keywords,
    expected,
):
    """Element with variadic arguments in signature."""
    model.add(fake_variadic_element_cls(name, *args, **kwargs))
    dump = next(
        iter(element_dump("fake_variadic_element", fake_variadic_element_cls, model))
    )
    script = unbuilder.unbuild(
        dump,
        argument_defaults=argument_defaults,
        prefer_keywords=prefer_keywords,
    )
    assert script == expected



import pytest
from finesse.script import KATSPEC, unparse


@pytest.mark.parametrize(
    "kwargs,directive_defaults,argument_defaults,prefer_keywords,expected",
    (
        # No lambda command.
        ({}, False, False, False, ""),
        ({}, False, False, True, ""),
        ({}, False, True, False, ""),
        ({}, False, True, True, ""),
        ({}, True, False, False, "lambda()"),
        ({}, True, False, True, "lambda()"),
        ({}, True, True, False, "lambda(1.064e-06)"),
        ({}, True, True, True, "lambda(value=1.064e-06)"),
        # 1550nm.
        ({"value": 1550e-9}, False, False, False, "lambda(1.55e-06)"),
        ({"value": 1550e-9}, False, False, True, "lambda(value=1.55e-06)"),
        ({"value": 1550e-9}, False, True, False, "lambda(1.55e-06)"),
        ({"value": 1550e-9}, False, True, True, "lambda(value=1.55e-06)"),
    ),
)
def test_lambda(
    model,
    kwargs,
    directive_defaults,
    argument_defaults,
    prefer_keywords,
    expected,
):
    adapter = KATSPEC.commands["lambda"]

    if kwargs:
        adapter.setter(model, ((), kwargs))

    dump = next(iter(adapter.getter(adapter, model)))
    script = unparse(
        dump,
        directive_defaults=directive_defaults,
        argument_defaults=argument_defaults,
        prefer_keywords=prefer_keywords,
    )
    assert script == expected



import pytest
import finesse
import numpy as np


@pytest.mark.parametrize("R", (0.9, 0.999))
@pytest.mark.parametrize("f", (100e3, 1e6, 100e6))
@pytest.mark.parametrize("fsig", (1e-12, 1, 10000))
@pytest.mark.parametrize("phase", (0, 33.4123, 90))
def test_rf_readout_vs_pd2_cavity_length(R, f, fsig, phase):
    """Tests that RF Redaout generates the same signal as a pd2 rf + fsig demodulation
    for a cavity length change."""
    model = finesse.Model()
    model.parse(
        f"""
    l l1
    mod mod1 f={f} midx=0.1 order=1 mod_type=pm
    m m1 R={R} T={1 - R}
    m m2 R=1 T=0
    link(l1, mod1, m1, 10, m2)

    readout_rf REFL9_X m1.p1.o f=mod1.f phase={phase}
    ad REFL9_X_I_signal REFL9_X.I f=fsig
    pd2 REFL9_X_I_signal_pd m1.p1.o f1=mod1.f phase1=REFL9_X.phase f2=fsig

    fsig({fsig})
    sgen sig m2.mech.z

    xaxis(m2.phi, lin, -1m, 1m, 100)
    """
    )

    out3 = model.run()
    pd2 = out3["REFL9_X_I_signal_pd"]
    rout = out3["REFL9_X_I_signal"]
    # this tends to zero error if steps -> inf
    assert max(abs(rout - pd2) / abs(rout)) < 1e-12


@pytest.mark.parametrize("R", (0.9, 0.999))
@pytest.mark.parametrize("f", (100e3, 1e6, 100e6))
@pytest.mark.parametrize("phase", (0, 33.4123, 90))
def test_rf_readout_vs_pd2_cavity_length_freq_sweep(R, f, phase):
    model = finesse.Model()
    model.parse(
        f"""
    l l1
    mod mod1 f={f} midx=0.1 order=1 mod_type=pm
    m m1 R={R} T={1 - R}
    m m2 R=1 T=0
    link(l1, mod1, m1, 10, m2)

    readout_rf REFL9_X m1.p1.o f=mod1.f phase={phase}
    ad REFL9_X_I_signal REFL9_X.I f=fsig
    pd2 REFL9_X_I_signal_pd m1.p1.o f1=mod1.f phase1=REFL9_X.phase f2=fsig

    fsig(1)
    sgen sig m2.mech.z

    xaxis(fsig, log, 1e-10, 1M, 100)
    """
    )

    out3 = model.run()
    assert (
        max(
            abs(out3["REFL9_X_I_signal_pd"] - out3["REFL9_X_I_signal"])
            / abs(out3["REFL9_X_I_signal_pd"])
        )
        < 1e-12
    )


@pytest.mark.parametrize("R", (0.9, 0.999))
@pytest.mark.parametrize("f", (100e3, 1e6, 100e6))
@pytest.mark.parametrize("phase", (0, 33.4123, 90))
def test_rf_readout_vs_pd2_cavity_length_tilt_single_element(R, f, phase):
    """Test that single element readout/pd2 do same calculations as not using a pdtype
    definition.

    Internally the calculations use different methods but should give identical results.
    """
    model = finesse.Model()
    model.parse(
        f"""
    l l1
    mod mod1 f={f} midx=0.1 order=1 mod_type=pm
    m m1 R={R} T={1 - R} Rc=-100
    m m2 R=1 T=0 Rc=100
    link(l1, mod1, m1, 10, m2)
    cav c m2.p1
    modes(maxtem=1)
    readout_rf REFL9_X m1.p1.o f=mod1.f phase={phase} output_detectors=True pdtype=single
    ad REFL9_X_I_signal REFL9_X.I f=fsig
    readout_rf REFL9_X2 m1.p1.o f=mod1.f phase={phase} output_detectors=True
    ad REFL9_X2_I_signal REFL9_X2.I f=fsig
    pd2 REFL9_X_I_signal_pd m1.p1.o f1=mod1.f phase1=REFL9_X.phase f2=fsig pdtype=single
    pd2 REFL9_X_I_signal_pd2 m1.p1.o f1=mod1.f phase1=REFL9_X.phase f2=fsig

    fsig(1M)
    sgen sig m2.mech.yaw

    xaxis(m2.xbeta, lin, -1u, 1u, 10)
    """
    )

    out3 = model.run()

    pd2 = out3["REFL9_X_I_signal_pd"]
    pd22 = out3["REFL9_X_I_signal_pd2"]
    rout = out3["REFL9_X_I_signal"]
    rout2 = out3["REFL9_X_I_signal"]
    assert max(abs(rout.real - pd2.real)) < 1e-10
    assert max(abs(rout.imag - pd2.imag)) < 1e-10
    assert max(abs(rout.real - pd22.real)) < 1e-10
    assert max(abs(rout.imag - pd22.imag)) < 1e-10
    assert max(abs(rout.real - rout2.real)) < 1e-10
    assert max(abs(rout.imag - rout2.imag)) < 1e-10

    # this tends to zero error if steps -> inf
    dx = out3.x1[1] - out3.x1[0]
    grad = np.gradient(out3["REFL9_X_I"], dx)
    assert max(abs(grad - pd2.real)) < 900000
    assert max(abs(grad - pd22.real)) < 900000


@pytest.mark.parametrize(
    ("pdtype", "node", "beta"),
    [
        ("xsplit", "yaw", "xbeta"),
        ("ysplit", "pitch", "ybeta"),
    ],
)
@pytest.mark.parametrize("phase", (0, 33.4123, 90))
def test_rf_readout_vs_pd2_cavity_tilt(pdtype, node, beta, phase):
    """Tests that readout gives the same error signal slopes as pd2 and DC response on
    reflection from a cavity when tilting mirrors."""
    model = finesse.Model()
    model.parse(
        f"""
    l l1
    mod mod1 f=1M midx=0.1 order=1 mod_type=pm
    m m1 R=0.99 T=0.01 Rc=-100
    m m2 R=1 T=0 Rc=100
    link(l1, mod1, m1, 10, m2)
    cav c m2.p1
    modes(maxtem=1)
    readout_rf REFL9_X m1.p1.o f=mod1.f phase={phase} output_detectors=True pdtype={pdtype}
    ad REFL9_X_I_signal REFL9_X.I f=fsig
    pd2 REFL9_X_I_signal_pd m1.p1.o f1=mod1.f phase1=REFL9_X.phase f2=fsig pdtype={pdtype}

    fsig(1M)
    sgen sig m2.mech.{node}

    xaxis(m2.{beta}, lin, -10u, 10u, 100)
    """
    )

    out3 = model.run()
    pd2 = out3["REFL9_X_I_signal_pd"]
    rout1 = out3["REFL9_X_I_signal"]
    assert max(abs(rout1.real - pd2.real)) < 1e-10
    assert max(abs(rout1.imag - pd2.imag)) < 1e-10

    # this tends to zero error if steps -> inf
    dx = out3.x1[1] - out3.x1[0]
    grad = np.gradient(out3["REFL9_X_I"], dx)
    assert max(abs(grad - pd2.real)) < 900000


@pytest.mark.parametrize("R", (0.9, 0.999))
@pytest.mark.parametrize("f", (100e3, 1e6, 100e6))
@pytest.mark.parametrize("phase", (0, 33.4123, 90))
@pytest.mark.parametrize("phi", (0, 0.2, -6.53))
def test_readout_rf_vs_pd1(R, f, phase, phi):
    model = finesse.Model()
    model.parse(
        f"""
    l l1
    mod mod1 f={f} midx=0.1 order=1 mod_type=pm
    m m1 R={R} T={1 - R}
    m m2 R=1 T=0 phi={phi}
    link(l1, mod1, m1, 10, m2)

    readout_rf REFL9_X m1.p1.o f=mod1.f phase={phase} output_detectors=True
    pd1 REFL9_X_I_pd1 m1.p1.o f=mod1.f phase=REFL9_X.phase

    xaxis(REFL9_X.phase, lin, -1, 1, 100)
    """
    )

    out = model.run()
    pd1 = out["REFL9_X_I_pd1"]
    rout = out["REFL9_X_I"]
    assert np.allclose(pd1, rout)


def test_readout_rf_default_f_with_fsig():
    """Tests the ReadoutRF runs without f being set, since f=None will break an fsig
    simulation."""
    model = finesse.Model()
    model.parse(
        """
    l l1
    readout_rf test l1.p1.o
    fsig(1)
    """
    )

    model.run()
    assert model.test.f.value is not None



"""Test cases for model component tree generation."""

import pytest


def test_component_tree_contains_network_nodes(network_model_michelson):
    """Test that all branches of the tree contain network nodes, regardless of starting
    node."""
    model, network = network_model_michelson

    for start in (model.L0, model.BS, model.IMX, model.IMY, model.EMX, model.EMY):
        tree = model.component_tree(start)
        assert tree.parent is None
        assert tree.name == start.name
        assert set([child.name for child in tree.get_all_children()]).issubset(
            set(network.nodes)
        )


@pytest.mark.parametrize(
    "root, radius, expected",
    (
        ("L0", 1, {"BS"}),
        ("L0", 2, {"BS", "IMX", "IMY"}),
        ("L0", 3, {"BS", "IMX", "IMY", "EMX", "EMY"}),
        ("BS", 1, {"L0", "IMX", "IMY"}),
    ),
)
def test_component_tree_radius(network_model_michelson, root, radius, expected):
    """Test that all branches of the tree contain network nodes, regardless of starting
    node."""
    model, network = network_model_michelson

    tree = model.component_tree(root=root, radius=radius)
    assert set([child.name for child in tree.get_all_children()]) == expected


def test_component_tree_flattens_cyclic_network(network_model_sagnac):
    """Test that component tree contains a flattened, acyclic version of a cyclic
    component network.

    Note: cyclic networks are handled by :meth:`networkx.dfs_tree`, so
    :meth:`.TreeNode.from_network` does not need to detect and avoid such cycles.
    """
    model, _ = network_model_sagnac

    for start in (model.L0, model.BS, model.M1, model.M2, model.M3):
        tree = model.component_tree(root=start)
        assert len(tree.get_all_children()) + 1 == len(model.components)


def test_component_tree_show_detectors(network_model_michelson):
    model, _ = network_model_michelson
    tree = model.component_tree(show_detectors=True)
    children = tree.get_all_children()
    names = [node.name for node in children]
    refl = children[names.index("refl")]
    assert refl.parent.name == "IMX"


@pytest.mark.parametrize(
    "show_detectors, show_ports, ref",
    (
        (
            False,
            False,
            # fmt:off
            (
                "○ L0\n"
                "╰──○ BS\n"
                "   ├──○ IMY\n"
                "   │  ╰──○ EMY\n"
                "   ╰──○ IMX\n"
                "      ╰──○ EMX"
            )
            # fmt:on
        ),
        (
            True,
            False,
            # fmt:off
            (
                "○ L0\n"
                "╰──○ BS\n"
                "   ├──○ IMY\n"
                "   │  ╰──○ EMY\n"
                "   ╰──○ IMX\n"
                "      ├──○ EMX\n"
                "      ╰──○ refl"
            )
            # fmt:on
        ),
        (
            False,
            True,
            # fmt:off
            (
                "○ L0\n"
                "╰──○ BS (L0.p1 ↔ BS.p1)\n"
                "   ├──○ IMY (BS.p2 ↔ IMY.p1)\n"
                "   │  ╰──○ EMY (IMY.p2 ↔ EMY.p1)\n"
                "   ╰──○ IMX (BS.p3 ↔ IMX.p1)\n"
                "      ╰──○ EMX (IMX.p2 ↔ EMX.p1)"
            )
            # fmt:on
        ),
        (
            True,
            True,
            # fmt:off
            (
                "○ L0\n"
                "╰──○ BS (L0.p1 ↔ BS.p1)\n"
                "   ├──○ IMY (BS.p2 ↔ IMY.p1)\n"
                "   │  ╰──○ EMY (IMY.p2 ↔ EMY.p1)\n"
                "   ╰──○ IMX (BS.p3 ↔ IMX.p1)\n"
                "      ├──○ EMX (IMX.p2 ↔ EMX.p1)\n"
                "      ╰──○ refl (IMX.p1.o)"
            )
            # fmt:on
        ),
    ),
)
def test_component_tree_draw(network_model_michelson, show_detectors, show_ports, ref):
    # can not use pregenerated component tree because we want to parameterize the args
    model, _ = network_model_michelson
    assert (
        model.component_tree(
            show_detectors=show_detectors, show_ports=show_ports
        ).draw_tree()
        == ref
    )
